%********************************************
%	OF_Best : OF of best indiviual and related data
%********************************************
%	01.03.00 - BR
%	22.03.00 - 
%********************************************

global	PrintTrace

global	one_third
global	epsilon_T
global	I									
global	J						
global	P					
global	K							
global	L						
global	M					
global	N				
global	O			
global	N_XT					
global	N_ST				
global	Stream			
global	HotUtility		
global	ColdUtility		
global	StorageFluid	
global	HEXCostFunction
global	HSUCostFunction	
global	T_Sp						
global	T_Tp				
global	X_SX
global	I_SX
global	T_min_HSU						
global	T_max_HSU						
global	Tk					
global	k_min				 
global	k_max			
global	T				
global	Optional_p	
global	Dt_batch					
global	Dtl					
global	RU				
global	Ut_X				
global	Abs_QH_CC				 
global	QC_HC				
global	MBB_k_fact	
global	MBB_kl_act
global	A_CP_h_po			
global	A_CP_h_pm			
global	A_CP_h_pn
global	TBCs_min
global	n_gen
%	global	Dvar
%********************************************

%	=====================================================================
%	================== ****** Start of SX_OF ****** =====================
%	=====================================================================

% *******************************************************
%	IHRS design and optimization using Genetic Algorithms
%
%	Overall SX Objective Function to be called by Struggle.
%
%	Receives the vector of independent variables X_SX of size
%	[1,K+P*(K-1)], converts them into temperatures, makes all
%	sizing and costs calculations and returns the value of the
%	objective function TBCs.
% *******************************************************
%	23.02.00 - last : xx.02.00 by PK
%	SX_OF.m
% *******************************************************

%	To be converted in a Matlab Function, and all process data stored in the Matlab
%	workspace which is used in this function be declared as global !!!!!!

%	Improvement opportunities: local copy of Utility and Storage Fluid parameters
%	to speed up the code ? (or do it in VarInit ?) e.g. TCU_in = ColdUtility(n).T_in, etc.

%	Note: the present implementation only holds for:
%		M = 1 , one type of hot utility
%		N = 1 , one type of cold utility
%		O = 1 , one type of storage fluid
%		N_XT = 1 , one type of HEX and associated cost function
%		N_ST = 1 , one type of HSU and associated cost function
%	Would require significant reading through and modifications
%	of the whole Matlab code, if these assumptions are not met.
%
%	PARAMETERS of "function" SX_OF
%
%	Input parameters supplied by Struggle:
%	**************************************
%	Vector X_SX of size [1,K+P*(K-1)] of independent variables:
%	the first K elements are converted in HSUs temperatures (vector
%	Tk [1,K]) allowing the calculation of the feasible range of k(p); 
%	these feasible ranges allow then to extract from
%	X_SX the meaningful variables (which are converted into T(p,k)
%	temperatures), while the other are skipped. X_SX is organized
%	in the following way: [HSUs variables ordered 1 ... K | I hot
%	streams variables each ordered K-1 ... 1 | J cold streams variables
%	each ordered 1 ... K-1]. Variables to be potentially skipped are 
%	those located at the beginning (on the left) of the set of
%	variables corresponding to a given process stream.
%
%	Note:
%	*****
%	the boundaries on variables (elements) of X_SX (Min [1,K+P*(K-1)]
%	and Max [1,K+P*(K-1)]) to be supplied to Struggle) are:
%		first K variables (to be decoded in Tk): epsilon_X <= X <= 1 - epsilon_X
%		remaining variables (to be decoded in T):	0 <= X <= 1 (epsilon_T is used during X -> T conversion "T_decoding")
%
%	Output parameters returned to Struggle:
%	***************************************
%	TBCs (total batch costs) is the objective function to be minimized.
%	Tk [1,K] contains the temperature of HSUs.
%	T [P,K] containing the "cut-off" temperatures of
%		process streams with each storage sub-system.
%	Additional parameters(?? to be declared as global ??):
%	IV_SX_act, the actual number of independent variables
%	k_min [1,P], k_max [1,P] to describe the range of variables of X_SX
%		which are actually meaningful
%	HR, the heat recovery
%	c_HEN, the costs of heat exchangers
%	c_STO, the costs of storage
%	c_UTIL, the costs of utilities
%	??????????????????????????????
%
%
%	Global variables from the Matlab workspace:
%	*******************************************
%	name, type, size, meaning, set in module ..., used in module ...
%	one_third							constant 1/3	set in "InitD", used for DTlm approx. of Chen
%	epsilon_T							constant for uniform boundaries on X variables for T(heat exchange) variables, set in "InitD", used in "T_Decoding" 
%	I										number of hot process streams, set in "DataInput", used as subscript/index boundary
%	J										number of cold process streams, set in "DataInput", used as subscript/index boundary
%	P										number of process streams (=I+J), set in "DataInput", used as subscript/index boundary
%	K										number of HSUs, set in "DataInput", used as subscript/index boundary
%	L										number of time slices, set in "GanttData", used as subscript/index boundary for mass balance in "HSUSizing"
%	M										number of hot utilities, set in "DataInput" (presently limited to 1)
%	N										number of cold utilities, set in "DataInput" (presently limited to 1)
%	O										number of storage fluids, set in "DataInput" (presently limited to 1)
%	N_XT									number of HEX types and cost functions, set in "DataInput" (presently limited to 1)
%	N_ST									number of HSU types and cost functions, set in "DataInput" (presently limited to 1)
%	Stream				[1,P]			structure of process streams data, set in "DataInput"
%	HotUtility			[1,M]			structure of hot utilities data, set in "DataInput"
%	ColdUtility			[1,N]			structure of cold utilities data, set in "DataInput"
%	StorageFluid		[1,O]			structure of storage fluid data, set in "DataInput"
%	HEXCostFunction	[1,N_XT]		structure of HEX cost functions data, set in "DataInput"
%	HSUCostFunction	[1,N_ST]		structure of HSU cost functions data, set in "DataInput"
%	T_Sp					[1,P]			supply temperatures of process streams, set in "VarInit"		
%	T_Tp					[1,P]			target temperatures of process streams, set in "VarInit"
%	X_SX				[1,K+P*(K-1)]	independent variables, set and modified by Struggle, used in "Tk_Decoding" and T_Decoding"
%	T_min_HSU							lower boundary of HSU operating temperature, set in "VarInit"
%	T_max_HSU							upper boundary of HSU operating temperature, set in "VarInit"
%	Tk						[1,K]			HSU temperatures, declared / initialized in "VarInitSX", set in "Tk_Decoding"
%	k_min					[1,P]			lower boundaries for meaningful X -> T decoding, declared / initialized in "VarInitSX", set "k_Range"
%	k_max					[1,P]			upper boundaries for meaningful X -> T decoding, declared / initialized in "VarInitSX", set "k_Range"
%	T						[P,K]			stream cut-off temperatures, declared / initialized in "VarInitSX", set "T_Decoding"
%	Optional_p			[1,P]			optional property of streams, set in "VarInit", =0 means process stream, =1 means optional heat recovery / soft stream
%	Dt_batch								batch cycle time (=(t_end-t_initial)/60 [h]), set in "GanttData"
%	Dtl					[1,L]			duration of time slices, set in "GanttData", used in "HSUSizing"
%	RU										reverse utility control parameter, set in "DataInput", mainly used in "T_Decoding" and "UtilitySizing", RU=0 reverse utilities not allowed, RU=1 reverse utilities allowed
%	Ut_X									utility HEX control parameter, set in "DataInput", used in "HEXCosts", 0<=Ut_X<=1, Ut_X=0 => costs of stility HEXs are not taken into account
%	Abs_QH_CC							total heating requirement of the process cold composite (i.e. accounting only for process streams), set in "DataInput", used 
%	QC_HC									total cooling need of the process hot composite (i.e. accounting only for process streams), set in "DataInput"
%	MBB_k_fact(p,o)	[P,O]			factors for initial mass imbalance calculation, set in "IndependentTerms", used in "HSUSizing"
%	MBB_kl_act(p,o,l)	[P,O,L]		factors (based on activity matrix) for slice-wise mass inventory calculation, set in "IndependentTerms", used in "HSUSizing"
%	A_CP_h_po			[P,O]			factors for HEX area calculation, set in "IndependentTerms", used in "HEXSizing"			
%	A_CP_h_pm			[P,M]			factors for HEX area calculation, set in "IndependentTerms", used in "UtilitySizing"
%	A_CP_h_pn			[P,N]			factors for HEX area calculation, set in "IndependentTerms", used in "UtilityHEXSizing"
%	TBCs_min								the current best (minimum) TBCs for display purpose
%	n_gen									the current equivalent number of generations
%
%	Local variables:
%	****************
%	????????? To avoid frequent reallocation, shouldn't we declare most of 
%	large matrices (e.g. MAB_kl, Atilda_pk ??? etc.) as global variables ??????????
%	any changes between local and global variables should be analysed with
%	respect to declaration / preallocation: e.g. k_max and k_min, which are
%	preallocated in VarInitSX but actually only used in SX_OF (this function)
%
%	iv_SX									subscript/index addressing the elements of X_SX, used in "T_Decoding"
%	i,j,p,k,l,n,m,o,n_XT,n_ST		subscript/index of related uppercase variable (refer to global variables)
%	DTK0,DTK1,DT1,DT2,DTpm,DTpn	various locally used temperature differences
%	HotTerm, ColdTerm					locally used factors (in "HSUSizing")
%	Current_A_CP_h_po					locally used factor (in "HEXSizing")
%	IV_SX_act							actual number of independent variables, set in "k_Range"
%	Atilda_pk			[P,K-1]		area of process-storage sub-system HEXs, set in "HEXSizing" 
%	AtildaMB_mk			[M,K-1] 		area of hot utility - storage mass balancing HEXs, set in "HSUSizing"
%	AtildaMB_nk			[N,K-1]			area of cold utility - storage mass balancing HEXs, set in "HSUSizing"
%	Atilda_pn			[P,N]			area of process-cold utility HEXs, set in "UtilitySizing"
%	Atilda_pm			[P,M]			area of process-hot utility HEXs, set in "UtilitySizing"	
%	DCl									duty cycle of a time slice, set and used in "HSUSizing"
%	DMBB_k				[1,K]			initial mass imbalance (after one batch cycle), calculated in "HSUSizing"
%	DM_k					[1,K]			cumulated mass imbalance (after one batch cycle), calculated in "HSUSizing"
%	QMB_mk				[M,K-1]		hot mass balancing utilities, calculated in "HSUSizing"
%	QMB_nk				[N,K-1]		cold mass balancing utilities, calculated in "HSUSizing"
%	QMB_m					[1,M]			required heating for mass balancing (summed on HSUs), calculated in "HSUSizing"
%	QMB_n					[1,N]			required cooling for mass balancing (summed on HSUs), calculated in "HSUSizing"
%	DMBB									before balance mass change of a HSU k during a time slice l, calculated in "HSUSizing"
%	MAB_kl				[K,L]			slice-wise, after balance mass inventories, calculated in "HSUSizing"
%	Min_MAB_k			[1,K]			minimum of slice-wise, after balance mass inventories, calculated in "HSUSizing"
%	Max_MAB_k			[1,K]			maximum of slice-wise, after balance mass inventories, calculated in "HSUSizing"
%	V_k					[1,K]			volume (capacity) of HSUs, calculated in "HSUSizing"
%	Atilda_p				[1,P]			area of process-storage HEXs (summed on Sss), calculated in "HEXCosts" 
%	c_X_p									total costs of individual process-storage HEXs on a process stream, calculated in "HEXCosts"
%	c_X_S_P				[P,1]			total costs of HEXs between a process stream and storage sub-systems, calculated in "HEXCosts"
%	c_X_S_P_tot							overall total costs of process-storage HEXs, calculated in "HEXCosts"
%	c_X_MB_mk			[M,K-1]		costs of individual mass balancing (hot utilities-Sss) HEXs, calculated in "HEXCosts"
%	c_X_MB_nk			[N,K-1]		costs of individual mass balancing (cold utilities-Sss) HEXs, calculated in "HEXCosts"
%	c_X_MB_tot							overall total costs of mass balancing (utilities-Sss) HEXs, calculated in "HEXCosts"
%	c_X_U_pm				[P,M]			costs of individual process-hot utilities HEXs, calculated in "HEXCosts"
%	c_X_U_pn				[P,N]			costs of individual process-cold utilities HEXs, calculated in "HEXCosts"
%	c_X_U_P_tot							overall total costs of process-utility HEXs, calculated in "HEXCosts"
%	Q_pm					[P,M] 		utility heating requirements at the end of each process stream, calculated in "UtilitySizing" !! be carefull to the reverse order of subscript !!
%	Q_pn					[P,N]      	utility cooling requirements at the end of each process stream, calculated in "UtilitySizing" !! be carefull to the reverse order of subscript !!
%	Q_m					[1,M]			utility heating requirements at the end of process streams (summed over process streams), calculated in "UtilityCosts"
%	Q_n					[1,N]			utility cooling requirements at the end of process streams (summed over process streams), calculated in "UtilityCosts"
%	THRH									total heat recovery heating between process streams (i.e. without any contribution of "optional" streams), calculated in "UtilityCosts" 
%	THRC									total heat recovery cooling between process streams (i.e. without any contribution of "optional" streams), calculated in "UtilityCosts"
%	HRS									total heating recovery from "optional" streams, calculated in "UtilityCosts"
%	Abs_CRS								total cooling recovery from "optional" streams, calculated in "UtilityCosts"
%	QH										total heating of hot utilities, calculated in "UtilityCosts"
%	QC										total cpoling of cold utilities, calculated in "UtilityCosts"
%	HR										total heating / cooling recovery of heat integration (i.e. process streams and optional streams), calculated in "UtilityCosts"
%	M_SFI									mass of storage fluid, calculated in "HSUCosts"	
%	c0X, C1X, mX, Amin, UnX			local copies of HEXCostFunction parameters, set in "HEXCosts"
%	c0S, c1S, mS, Vmin				local copies of HSUCostFunction parameters, set in "HSUCosts"
%	c_HEN									total costs of HEXs [CHF/batch], calculated in "HEXCosts"
%	c_HSU									total costs of HSUs [CHF/batch], calculated in "HSUCosts"
%	c_SFI									costs of storage fluid inventory [CHF/batch], calculated in "HSUCosts"
%	c_STO									total costs of heat storage (including storage tanks and fluid inventory) [CHF/batch], calculated in "HSUCosts"
%	c_UTIL								total costs of utilities, calculated in "UtilityCosts" [CHF/batch]
%	c_UTIL_heating						total costs of hot utilities, calculated in "UtilityCosts" [CHF/batch]
%	c_UTIL_cooling						total costs of cold utilities, calculated in "UtilityCosts" [CHF/batch]
%	TBCs									total batch costs, calculated in "TotalBatchCosts" [CHF/batch]
%
%
%	!!!!!!!! Declaration of global variables !!!!!!!!!
%
%	Preallocation (could be made global variables ???) / but INITIALIZATION TO ZERO IS ESSENTIAL !!!!!
m = 1;
n = 1;
o = 1;
n_XT = 1;
n_ST = 1;
%	for k_Range
k_max(1:I) = k_min(1:I);
k_min(I+1:P) = k_max(I+1:P);
%	for T_Decoding
for p = 1:P
   T(p,:) = T_Sp(p);
end
%	for HEXSizing
Atilda_pk = zeros(P,K-1);
%	for HSUSizing
DMBB_k 		= zeros(1,K);
DM_k			= zeros(1,K);
QMB_mk		= zeros(M,K-1);
QMB_nk		= zeros(N,K-1);
AtildaMB_mk		= zeros(M,K-1);
AtildaMB_nk		= zeros(N,K-1);
MAB_kl 		= zeros(K,L);
Min_MAB_k		= zeros(1,K);
Max_MAB_k 		= zeros(1,K);
V_k 			= zeros(1,K);
%	for UtilitySizing
Q_pm      		= zeros(P,M);
Q_pn      		= zeros(P,N);
Atilda_pm 		= zeros(P,M);
Atilda_pn 		= zeros(P,N);
%	for HEXCosts
Atilda_p 		= zeros(1,P);
c_X_S_P		= zeros(1,P);
c_X_MB_mk		= zeros(M,K-1);
c_X_MB_nk		= zeros(N,K-1);
c_X_U_pm		= zeros(P,M);
c_X_U_pn		= zeros(P,N);
%	for HSUCosts
%
%	for UtilityCosts
Q_m			= zeros(1,M);
Q_n			= zeros(1,N);
QMB_m			= zeros(1,M);
QMB_n			= zeros(1,N);

%	=====================================================================
%	==>	Begin of Tk_Decoding
for k=1:K
	if k > 1
		Tk(k)   = (1 - X_SX(k)) * Tk(k-1)   + X_SX(k) * T_max_HSU;
	else
		Tk(k)   = (1 - X_SX(k)) * T_min_HSU + X_SX(k) * T_max_HSU;
	end
end
%	<==	End of Tk_Decoding
%	=====================================================================
%	==>	Begin of Reading Xp
if Binary_var_mode == 1	%	else, although generated in any case, the Xp variables are ignored and Xp remains unchanged after initialization at 1
	for p=1:P
      Xp(p)= round(X_SX(p+K));
   end
end
%	<==	End of Reading Xp
%	=====================================================================
%	==>	Begin of k_Range
IV_SX_act = 0;
for p = 1:I
   if (Xp(p) ~= 0)	%	otherwise, the optimization has decided not to place a match on this stream
		for k = 2:K
			if T_Sp(p) > Tk(k)
				k_max(p) = k;
			end
   	end
   	if k_max(p) >= 2
      	IV_SX_act = IV_SX_act + k_max(p) - 1;
      end
   end
end
for p = I+1:P
   if (Xp(p) ~= 0)	%	otherwise, the optimization has decided not to place a match on this stream
		for k = K-1:-1:1
			if T_Sp(p) < Tk(k) 
				k_min(p) = k;
			end
		end
   	if k_min(p) <= K-1
         IV_SX_act = IV_SX_act + K - k_min(p);
      end
   end
end
%	<==	End of k_Range
%	=====================================================================
%	==>	Begin of T_Decoding
%	Temperature of hot streams
iv_SX = K+P;
for p=1:I
   iv_SX = iv_SX + K-k_max(p);
   if k_min(p) < k_max(p)
      for k=k_max(p)-1:-1:k_min(p)
			iv_SX = iv_SX + 1;
			if (RU == 0) & (Tk(k) < T_Tp(p))
      		T(p,k) = (1 - X_SX(iv_SX)) * T(p,k+1) + X_SX(iv_SX) * T_Tp(p);		
         else
      		T(p,k) = (1 - X_SX(iv_SX)) * T(p,k+1) + X_SX(iv_SX) * (Tk(k) + epsilon_T);		
         end
   	end
   end
end	%	iv_SX pointing on last X variable of last hot stream
% 	Temperature of cold streams
for p=I+1:P
   iv_SX = iv_SX + k_min(p)-1;
   if k_min(p) < k_max(p)
      for k=k_min(p):k_max(p)-1
			iv_SX = iv_SX + 1;
         if (RU == 0) & (Tk(k+1) > T_Tp(p))
				T(p,k+1) = (1 - X_SX(iv_SX)) * T(p,k) + X_SX(iv_SX) * T_Tp(p);
         else
				T(p,k+1) = (1 - X_SX(iv_SX)) * T(p,k) + X_SX(iv_SX) * (Tk(k+1) - epsilon_T);
         end
      end
   end
end	%	iv_SX pointing on last X variable of last cold stream
%	<==	End of T_Decoding
%	=====================================================================
%	==>	Begin of HEXSizing
o = 1;
for p = 1:P
   Current_A_CP_h_po = A_CP_h_po(p,o);
   if k_min(p) < k_max(p)
      for k = k_min(p):k_max(p)-1
         DT1 = T(p,k+1) - Tk(k+1);
			DT2 = T(p,k)   - Tk(k);
         Atilda_pk(p,k) = Current_A_CP_h_po * (T(p,k+1) - T(p,k)) / ((abs(0.5 * DT1 * DT2 * (DT1 + DT2))) ^ one_third);
      end
	end
end
%	<==	End of HEXSizing
%	=====================================================================
%	==>	Begin of HSUSizing
%	Overall mass imbalance (after one batch cycle)
k = 1;
for p = 1:P
  DMBB_k(k) = DMBB_k(k) + MBB_k_fact(p,o) * ( - ((T(p,k+1) - T(p,k)) / (Tk(k+1) - Tk(k))));
end
if K > 2
   for k = 2:K-1
		DTK1 = Tk(k+1) - Tk(k);
   	DTK0 = Tk(k) - Tk(k-1);
   	for p = 1:P
   		DMBB_k(k) = DMBB_k(k) + MBB_k_fact(p,o) * (((T(p,k) - T(p,k-1)) / DTK0) - ((T(p,k+1) - T(p,k)) / DTK1));
      end
   end
end
k = K;
for p = 1:P
   DMBB_k(k) = DMBB_k(k) + MBB_k_fact(p,o) * (T(p,k) - T(p,k-1)) / (Tk(k) - Tk(k-1));
end
%	Cumulated mass imbalance (after one batch cycle)
DM_k(1) = DMBB_k(1);
for k = 2:K
   DM_k(k) = DM_k(k-1) + DMBB_k(k);
end
%	Required mass balancing utilities & MB HEX areas
HotTerm = ((1 / HotUtility(m).h) + (1 / StorageFluid(o).h_SF)) / Dt_batch; 
ColdTerm = ((1 / ColdUtility(n).h) + (1 / StorageFluid(o).h_SF)) / Dt_batch;
for k = 1:K-1
   if DM_k(k)> 0
      QMB_mk(m,k) = DM_k(k) * StorageFluid(o).cp_SF * (Tk(k+1) - Tk(k)) / 3600;
      DT1 = HotUtility(m).T_in - Tk(k+1);
      DT2 = HotUtility(m).T_out - Tk(k);
      AtildaMB_mk(m,k) = QMB_mk(m,k) * HotTerm / ((0.5 * DT2 * DT1 * (DT2 + DT1)) ^ one_third);
   else
      QMB_nk(n,k) = DM_k(k) * StorageFluid(o).cp_SF * (Tk(k) - Tk(k+1)) / 3600;
      DT1 = Tk(k+1) - ColdUtility(n).T_out;
      DT2 = Tk(k) - ColdUtility(n).T_in;
      AtildaMB_nk(n,k) = QMB_nk(n,k) * ColdTerm / ((0.5 * DT2 * DT1 * (DT2 + DT1)) ^ one_third);
   end
end
%	Slice-wise mass inventory and required capacities
l = 1;
DCl = Dtl(l) / Dt_batch;
k = 1;
DMBB = 0;
for p = 1:P
   DMBB = DMBB + MBB_kl_act(p,o,l) * ( - ((T(p,k+1) - T(p,k)) / (Tk(k+1) - Tk(k))));
end
MAB_kl(k,l) = DMBB - DCl * DMBB_k(k);
if MAB_kl(k,l) > Max_MAB_k(k)
   Max_MAB_k(k) = MAB_kl(k,l);
else
   if MAB_kl(k,l) < Min_MAB_k(k)
   	Min_MAB_k(k) = MAB_kl(k,l);
   end
end
if K > 2
   for k = 2:K-1
   	DTK1 = Tk(k+1) - Tk(k);
   	DTK0 = Tk(k) - Tk(k-1);
      DMBB = 0;
      for p = 1:P
         DMBB = DMBB + MBB_kl_act(p,o,l) * (((T(p,k) - T(p,k-1)) / DTK0) - ((T(p,k+1) - T(p,k)) / DTK1));
      end
      MAB_kl(k,l) = DMBB - DCl * DMBB_k(k);
      if MAB_kl(k,l) > Max_MAB_k(k)
         Max_MAB_k(k) = MAB_kl(k,l);
      else
         if MAB_kl(k,l) < Min_MAB_k(k)
            Min_MAB_k(k) = MAB_kl(k,l);
         end
      end
   end
end
k = K;
DMBB = 0;
for p = 1:P
   DMBB = DMBB + MBB_kl_act(p,o,l) * (T(p,k) - T(p,k-1)) / (Tk(k) - Tk(k-1));
end
MAB_kl(k,l) = DMBB - DCl * DMBB_k(k);
if MAB_kl(k,l) > Max_MAB_k(k)
   Max_MAB_k(k) = MAB_kl(k,l);
else
   if MAB_kl(k,l) < Min_MAB_k(k)
      Min_MAB_k(k) = MAB_kl(k,l);
   end
end
% end of particular case l=1 (avoids multiple tests if l=1 ...)
for l = 2:L
   DCl = Dtl(l) / Dt_batch;
	k = 1;
	DMBB = 0;
	for p = 1:P
	   DMBB = DMBB + MBB_kl_act(p,o,l) * ( - ((T(p,k+1) - T(p,k)) / (Tk(k+1) - Tk(k))));
	end
   MAB_kl(k,l) = MAB_kl(k,l-1) + DMBB - DCl * DMBB_k(k);
	if MAB_kl(k,l) > Max_MAB_k(k)
	   Max_MAB_k(k) = MAB_kl(k,l);
	else
 	   if MAB_kl(k,l) < Min_MAB_k(k)
   		Min_MAB_k(k) = MAB_kl(k,l);
  	 	end
	end
	if K > 2
   	for k = 2:K-1
   		DTK1 = Tk(k+1) - Tk(k);
   		DTK0 = Tk(k) - Tk(k-1);
      	DMBB = 0;
      	for p = 1:P
         	DMBB = DMBB + MBB_kl_act(p,o,l) * (((T(p,k) - T(p,k-1)) / DTK0) - ((T(p,k+1) - T(p,k)) / DTK1));
      	end
         MAB_kl(k,l) = MAB_kl(k,l-1) + DMBB - DCl * DMBB_k(k);
      	if MAB_kl(k,l) > Max_MAB_k(k)
         	Max_MAB_k(k) = MAB_kl(k,l);
      	else
         	if MAB_kl(k,l) < Min_MAB_k(k)
            Min_MAB_k(k) = MAB_kl(k,l);
         	end
      	end
   	end
	end
	k = K;
	DMBB = 0;
	for p = 1:P
   	DMBB = DMBB + MBB_kl_act(p,o,l) * (T(p,k) - T(p,k-1)) / (Tk(k) - Tk(k-1));
	end
   MAB_kl(k,l) = MAB_kl(k,l-1) + DMBB - DCl * DMBB_k(k);
	if MAB_kl(k,l) > Max_MAB_k(k)
   	Max_MAB_k(k) = MAB_kl(k,l);
	else
   	if MAB_kl(k,l) < Min_MAB_k(k)
      	Min_MAB_k(k) = MAB_kl(k,l);
   	end
   end
end
%	Required capacity of HSUs
for k = 1:K
   V_k(k) = (Max_MAB_k(k) - Min_MAB_k(k)) / StorageFluid(o).rho_SF;
end
%	<==	End of HSUSizing
%	=====================================================================
%	==>	Begin of UtilitySizing
%	Utility (normally cold, but may also be hot if RU=1) needed at the end of hot process streams
k=1;
for p = 1:I
   if (Optional_p(p) == 0) & (T(p,k) ~= T_Tp(p))
      if T(p,k) > T_Tp(p)
         DTpn = T(p,k)  - T_Tp(p);
			DT1  = T(p,k)  - ColdUtility(n).T_out;
			DT2  = T_Tp(p)	- ColdUtility(n).T_in;
			Q_pn(p,n) = Stream(p).CP_Dt * DTpn;
         Atilda_pn(p,n) = A_CP_h_pn(p,n) * DTpn / (( 0.5 * DT1 * DT2 * ( DT1 + DT2 ) ) ^ one_third);
		else	 
			DTpm = T_Tp(p)   				- T(p,k);
			DT1  = HotUtility(m).T_in  - T_Tp(p);
			DT2  = HotUtility(m).T_out - T(p,k);
			Q_pm(p,m) = Stream(p).CP_Dt * DTpm;
         Atilda_pm(p,m) = A_CP_h_pm(p,m) * DTpm / (( 0.5 * DT1 * DT2 * ( DT1 + DT2 ) ) ^ one_third);
		end
	end
end
%	Utility (normally hot, but may also be cold if RU=1) needed at the end of cold process streams
k=K;
for p = I+1:P
	if (Optional_p(p) == 0) & (T(p,k) ~= T_Tp(p))
		if T(p,K) < T_Tp(p)
			DTpm = T(p,k)             - T_Tp(p);
			DT1  = HotUtility(m).T_in  - T_Tp(p);
			DT2  = HotUtility(m).T_out - T(p,k);
			Q_pm(p,m) = Stream(p).CP_Dt * DTpm;
			Atilda_pm(p,m) = A_CP_h_pm(p,m) * DTpm / (( 0.5 * DT1 * DT2 * ( DT1 + DT2 ) ) ^ one_third);
		else
			DTpn = T_Tp(p)	- T(p,k);
			DT1  = T(p,k) - ColdUtility(n).T_out;
			DT2  = T_Tp(p)	- ColdUtility(n).T_in;
			Q_pn(p,n) = Stream(p).CP_Dt * DTpn;
         Atilda_pn(p,n) = A_CP_h_pn(p,n) * DTpn / (( 0.5 * DT1 * DT2 * ( DT1 + DT2 ) ) ^ one_third);
		end
	end
end
%	<==	End of UtilitySizing
%	=====================================================================
%	==>	Begin of HEXCosts
%	local copies of frequently used data, likely to speed up calculations
c0X 	= HEXCostFunction(n_XT).c0_X;
c1X	= HEXCostFunction(n_XT).c1_X;
mX 	= HEXCostFunction(n_XT).m_X;
Amin 	= HEXCostFunction(n_XT).A_min;
UnX 	= HEXCostFunction(n_XT).Un_X;
%	Costs of process-Storage HEXs
Atilda_p = sum(Atilda_pk,2);
   
for p = 1:P
   c_X_p = 0;
   for k = 1:K-1
      c_X_p = c_X_p + ( c0X + c1X * Atilda_pk(p,k) ^ mX ) * ( 1-exp(-5*Atilda_pk(p,k) / Amin));
   end
   c_X_S_P(p) = (1 - UnX) * ( c0X + c1X * Atilda_p(p) ^ mX ) * ( 1-exp(-5*Atilda_p(p) / Amin)) + UnX * c_X_p;
end
c_X_S_P_tot = sum(c_X_S_P);
%	Costs of mass-balancing HEXs for HSUs	!!! should be summed over m and n for multiple utilities
c_X_MB_tot = 0;
for k = 1:K-1
   c_X_MB_mk(m,k) = (c0X + c1X * AtildaMB_mk(m,k) ^ mX ) * ( 1-exp(-5*AtildaMB_mk(m,k) / Amin));
   c_X_MB_nk(n,k) = (c0X + c1X * AtildaMB_nk(n,k) ^ mX ) * ( 1-exp(-5*AtildaMB_nk(n,k) / Amin));
end
c_X_MB_tot = sum(sum((c_X_MB_mk)')) + sum(sum((c_X_MB_nk)'));
%	Costs of utility HEXs on process streams	!!! should be summed over m and n for multiple utilities
for p= 1:P
   c_X_U_pm(p,m) = (c0X + c1X * Atilda_pm(p,m) ^ mX ) * ( 1-exp(-5*Atilda_pm(p,m) / Amin));
   c_X_U_pn(p,n) = (c0X + c1X * Atilda_pn(p,n) ^ mX ) * ( 1-exp(-5*Atilda_pn(p,n) / Amin));
end
c_X_U_P_tot = sum(sum(c_X_U_pm)) + sum(sum(c_X_U_pn));
%	Total costs of HEXs
c_HEN = c_X_S_P_tot + c_X_MB_tot + Ut_X * c_X_U_P_tot;
%	<==	End of HEXCosts
%	=====================================================================
%	==>	Begin of HSUCosts
%	local copies of frequently used data, likely to speed up calculations
c0S 	= HSUCostFunction(n_ST).c0_S;
c1S	= HSUCostFunction(n_ST).c1_S;
mS 	= HSUCostFunction(n_ST).m_S;
Vmin 	= HSUCostFunction(n_ST).V_min;
c_HSU = 0;
for k = 1:K
	c_HSU = c_HSU + ( c0S + c1S * V_k(k) ^ mS ) * ( 1 - exp ( -5 * V_k(k) / Vmin ) );
end
M_SFI = - sum(Min_MAB_k);
c_STO = c_HSU + StorageFluid(o).c_SF * ( M_SFI / StorageFluid(o).rho_SF );
%	<==	End of HSUCosts
%	=====================================================================
%	==>	Begin of UtilityCosts
%	Heating / cooling recovery from "optional" streams
HRS = 0;
Abs_CRS = 0;
for p = 1:I
   if Optional_p(p) == 1
      HRS = HRS + Stream(p).CP_Dt * (T_Sp(p) - T(p,1));
   end
end
for p = I+1:P
   if Optional_p(p) == 1
      Abs_CRS = Abs_CRS + Stream(p).CP_Dt * (T_Sp(p) - T(p,K));
   end
end
%	Heating / cooling requirements of process streams (summed on P streams)
Q_m = sum(Q_pm);
Q_n = sum(Q_pn);
%	Heating / cooling requirements for mass-balancing HSUs (summed on K-1 Sss)
QMB_m = sum((QMB_mk)');
QMB_n = sum((QMB_nk)');
%	Summing heating / cooling
QH = sum(Q_m) + sum(QMB_m); 
QC = sum(Q_n) + sum(QMB_n);
THRH = Abs_QH_CC - QH - HRS;
THRC = QC_HC - QC - Abs_CRS;
HR = THRH + HRS + Abs_CRS;
%	Utility costs
c_UTIL_heating = HotUtility(m).c * QH;
c_UTIL_cooling = ColdUtility(n).c * QC;
c_UTIL = c_UTIL_heating + c_UTIL_cooling;
%	<==	End of UtilityCosts
%	=====================================================================
%	==>	Begin of TotalBatchCosts
TBCs = c_HEN + c_STO + c_UTIL;
%	<==	End of TotalBatchCosts

%	Data to be displayed:

X_SX
dispnum ( TBCs, 'TBCs');
dispnum ( HR, 'HR');
dispnum ( c_HEN, 'c_HEN');
Atilda_pk
Atilda_p
AtildaMB_mk
AtildaMB_nk
Atilda_pn
Atilda_pm
dispnum ( c_STO, 'c_STO');
V_k
DMBB_k
dispnum (M_SFI, 'M_SFI');
dispnum ( c_UTIL, 'c_UTIL');
dispnum ( QH, 'QH');
dispnum ( QC, 'QC');
dispnum ( QMB_m, 'QMB_m');
dispnum ( QMB_n, 'QMB_n');
dispnum ( Q_m, 'Q_m');
dispnum ( Q_n, 'Q_n');
Tk
T
k_min
k_max
Xp
%	=====================================================================
%	=================== ****** End of SX_OF ****** ======================
%	=====================================================================
