/* ----------------------------------------------------------------------------
  FOpt.c

  Example program for the use of MLGAMain module

  BR K316 - K429
  Optimize v3.92 - K524
  Optimize v4.12 - Ka27
---------------------------------------------------------------------------- */

#include <stdio.h>
#include <cmath>
#include <string.h>
#include "MLGAMain.h"
#include "mlio.h"

#define SQR(x) ((x)*(x))

///////////////////////////////////////////////////////
// Main entry point from mathlab
///////////////////////////////////////////////////////
void mexFunction(int nlhs,
				 mxArray *plhs[], 
				 int nrhs,
				 const mxArray *prhs[] )
{
	MLGAMain	GAMain;

	GAMain.Main(nlhs, plhs, nrhs, prhs);

}

/////////////////////////////////////////////////////
// Simulation de _OF_ (Objective function)
/////////////////////////////////////////////////////
double simul_OF_ (double *CurrentGenome,	// Genome
					int TailleGenome,		// Genome size
					GAGenome& g)
{
    int     i;
	int		GTkGTsize;
 	int     n_out, n_in;
	mxArray *oa[3], *ia[3];
	double  *GTkGT, *o_OFset, obj;
	int		nrow0,  ncol0;

	if (MLIO::PrintTrace() > 1)	
		printf( "\n\n simul_OF_ : Call of OFm : CurrentGenome[1] = %f - size genome = %d\n", 
				CurrentGenome[1], TailleGenome );

	n_out = 3;
	n_in  = 3;

	// Input array initialization
	ia[0] = mxCreateString( "Genome from simul_OF_ !" );
	ia[1] = mxCreateDoubleMatrix( 1, TailleGenome, mxREAL);
	memcpy( mxGetPr(ia[1]), CurrentGenome, TailleGenome*sizeof(double) );
	ia[2] = mxCreateDoubleMatrix(1, 1, mxREAL);
	*mxGetPr(ia[2]) = TailleGenome;

	if (MLIO::PrintTrace() > 1)	
		printf ( " Just before mexCallMATLAB(n_out, oa, n_in, ia, ""ROFm"" )\n" );
	
	// Call mathlab Objective function
	mexCallMATLAB(n_out, oa, n_in, ia, "ROFm" );
	
	if (MLIO::PrintTrace() > 1)	
		printf ( " Just after  mexCallMATLAB(n_out, oa, n_in, ia, ""ROFm"" )\n\n" );

	o_OFset = mxGetPr(oa[0]);    //	c_HEN, c_STO, c_UTIL, TBCs, AveTBC, HR - 6 SideFun

	// with GTkGT
	GTkGT = (double *) mxGetPr(oa[1]);		//	Super vector, size Dsize
	GTkGTsize = (int) mxGetScalar(oa[2]);

	MLEvalData *EVDT;
	EVDT = ((MLEvalData *) (g.evalData()));
	if (EVDT == NULL)
		printf ( " (MLEvalData *) (g.evalData()) == NULL   $$\n");

	EVDT->SetGenome(GTkGTsize,GTkGT);

	nrow0 = mxGetM(oa[0]);
	if (nrow0 == 1)
		{
		ncol0 = mxGetN(oa[0]);
		}
		else
		{
		printf ( " ERROR : Return parameter nrow0 is not a vector = mxGetM(oa[0]) = %d\n", nrow0);
		return 999999;
		}
	EVDT->SetGenomeValue(GTkGTsize,(double) (ncol0));	// Number of side functions
	for ( i = 0; i < ncol0; i++ )
		{
		EVDT->SetGenomeValue(GTkGTsize+i+1,o_OFset[i]);
		}
	obj = o_OFset[3];			// Yes, it's the forth one !

	/*****************  release memory  *********************/
	for( i = 0; i < 3; i++ )
		{
		mxDestroyArray( oa[i] );
		mxDestroyArray( ia[i] );
		}

	return obj;

} // simul_OF_


/////////////////////////////////////////////////////
// Initialisation de _OF_ (Objective function)
/////////////////////////////////////////////////////
void init_OF_ (double *CurrentGenome,
			   double *InitialGenome, 
			   int TailleGenome)
{
	memcpy ( CurrentGenome, InitialGenome, TailleGenome*sizeof(double) );
} // init_OF


//////////////////////////////////////////////////////////
// simple comparator with Temperatures and Weights
//////////////////////////////////////////////////////////
float my_comparator(const GAGenome& g1, 
					const GAGenome& g2) 
{
//	return 1;

	MixedGenome		&a = ( MixedGenome &) g1;
	MixedGenome		&b = ( MixedGenome &) g2;
	float			sum=0;
	int				m, n;
	
	MLInterface		*d = (MLInterface *)a.userData();// my population data structure
	double			Dsize = d->ReadDouble(sDsize);
	double			*inDwei = d->ReadArray(sDwei, &m, &n);
	
	int k;
	static int Iloop = 0;
	double TTka[300], TTkb[300];
	double *PtrTTka = TTka;
	double *PtrTTkb = TTkb;

	Iloop += 1;

	/**********************************************************************************
		Aim here is to reread and print the temperatures GTk and GT
		by GetTemperatureArray of MLevalData */

	MLEvalData *a_evda;
	MLEvalData *b_evda;
	a_evda = ((MLEvalData *) (a.evalData()));	// my individual data structure
	b_evda = ((MLEvalData *) (b.evalData()));
	if (a_evda == NULL) 
		printf ( " ((MLEvalData *) (a.evalData())) == NULL   $$\n");
	if (b_evda == NULL) 
		printf ( " ((MLEvalData *) (b.evalData())) == NULL   $$\n");

	PtrTTka = a_evda->GetGenome();
	PtrTTkb = b_evda->GetGenome();

	if (MLIO::PrintTrace() > 1)	
		printf( "**************************** Calculate distance *******************************\n" );

	for (k=0;k<Dsize;k++)
		{
		sum += SQR( inDwei[k] * ( PtrTTka[k]-PtrTTkb[k] ) );	// p = 0 : Dsize-1
		}

	if (MLIO::PrintTrace() > 1)
		printf ( "  Iloop = %d * Distance using Super Temperature Array : sum = %9.5lf\n",Iloop,sum );
	if (MLIO::PrintTrace() > 1)	
		printf( "*********************************** Done **************************************\n\n" );

	return sqrt(sum);
} // my_Comparator


//////////////////////////////////////////////////////////
// Calculate objective function
//////////////////////////////////////////////////////////
float CalcObjective(GAGenome& g) 
{
	MixedGenome& genome = (MixedGenome&)g;
	static double	*x = NULL;
	static double	*kvals=NULL; 
	double			obj=0.0;
	int				countreal,countint,i;

	// Read user data
	MLInterface		*d = (MLInterface *)genome.userData();	// my data structure
	int				Xsize = (int)d->ReadDouble(sXsize);
	double			*inXval = d->ReadArray(sXval);
	double			*inXopt = d->ReadArray(sXopt);

	if (MLIO::PrintTrace() > 0)	printf("*************\nStart CalcObjective(GAGenome& g) : Call objective function : size=%li, val=%lf, opt=%lf\n", Xsize,inXval[1],inXopt[1]); // Curieux %li ?

/*
	d->PrintArg();
*/
	// Init if it's the first time
	if(!x) 
		{
		x= new double[Xsize];
		init_OF_(x,inXval,Xsize);
		}

	if (MLIO::PrintTrace() > 1)	printf("0=* write for x[], init_OF_ XSize = %d\n", Xsize);
	countreal=0;
	countint =0;
	for(i=0;i<Xsize;++i) 
		{
		switch ((int)(10*inXopt[i]))
			{
			case 10 :
			    // this parameter is in the integer gene to be optimized
	if (MLIO::PrintTrace() > 1)	printf("1=* write in x[%2d] int  from gene()   = %lf\n",i,genome.int_genome().gene(countint));
				x[i]=genome.int_genome().gene(countint);
				countint++;
				break;
			case 9 :
			    // this parameter is in the real    gene to be optimized
	if (MLIO::PrintTrace() > 1)	printf("2=* write in x[%2d] real from gene()   = %lf\n",i,genome.real_genome().gene(countreal));
				x[i]=genome.real_genome().gene(countreal);
				countreal++;
				break;
			case 1 :
				// it's not, so find it from the original real    data
	if (MLIO::PrintTrace() > 1)	printf("3=* write in x[%2d] real from inXval() = %lf\n",i,inXval[i]);
				x[i]=inXval[i];			
				break;
			case 0 :
				// it's not, so find it from the original integer data
	if (MLIO::PrintTrace() > 1)	printf("4=* write in x[%2d] int  from inXval() = %lf\n",i,inXval[i]);
				x[i]=inXval[i];			
				break;
			default :
	printf ( "Wrong value for Xopt (should be 0.0, 0.1, 0.9 or 1.0)\tinXopt[%d]=%lf\n\n ***STOP ! ***\n\n",i,inXopt[i] );
				return(990099);
			}
		}	//	for
	if (MLIO::PrintTrace() > 1)	
		{
		printf( "* Middle part of CalcObjective, before simul_of_  ! ******* ???????????????? ******\n" );
		printf (" x[1:%d] = ", Xsize );
		for(i=0;i<Xsize;++i) 
			{
			printf ( " %5.3f",x[i] );
			}
		} // if

	obj = simul_OF_(x,Xsize,g);

#ifdef PRINT_VALUES
	for(i=0;i<Xsize;++i) 
		{
		BestIndOfGen<<"x["<<i<<"]="<<x[i]<<"\n";
		}
	BestIndOfGen<<"obj="<<obj<<"\n";
#endif

	if (MLIO::PrintTrace() > 0)	printf("End  CalcObjective(GAGenome& g)\n\n");
  return ((float)obj);
} // CalcObjective
