//////////////////////////////////////////////////////////////////////
// MLGAMain.cpp: implementation of the MLGAMain class.
//
// see MLGAMain.h for explanation header
//
//  Optimize v4.12 - Ka27
//////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include "StruggleH.h"
#include <cmath>
#include <string.h>
// Definition of Genome private data
#include "MLGAMain.h"
#include "mlio.h"


#define SETTINGS_FILE "settings.txt"
const int	SideFunDim = 6;


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MLGAMain::MLGAMain()
{

}

MLGAMain::~MLGAMain()
{

}

///////////////////////////////////////////////////////
// Main entry point from MATLAB
///////////////////////////////////////////////////////
void MLGAMain::Main(int nlhs,
				 mxArray *plhs[], 
				 int nrhs,
				 const mxArray *prhs[] )
{
	MLInterface UserData;
	MLInterface Parametres;
	MLInterface ControlData;
	
	const int	dims = 1;
	int		inXsize, inDsize, inPTra, inPnPlot;
	int		i, Xm, Xn, Dm, Dn;
	double	OptOBJfun, SideFunNum;
	double  SideFun[SideFunDim];
	mxArray	*array_ptr;

	int Taille = dims;
	int VirtualTable[dims]={SideFunDim};		//virtual table with one element : 6 for mxCreateNumericArray

	mexPrintf( "\n   Start MEX interface optimize.dll v4.20 (MATLAB <-> VC++) - 27.12.2000  \n" );
    // verify in/out arguments
	if( nlhs != 3 )
		mexErrMsgTxt("Output must be Genome, Result, FOpt(Result)\n");
	if( nrhs != 3 )
		mexErrMsgTxt("Input must be UserData, Parameters, ControlData\n");

	UserData.SetArgs(prhs[0], true);
	UserData.PrintArg();
	Parametres.SetArgs(prhs[1]);
	Parametres.PrintArg();
	ControlData.SetArgs(prhs[2]);
	ControlData.PrintArg();
	
	mexPrintf("MLGAMain::Main ----> Trace = %i\n", MLIO::PrintTrace());

	/* Unto here */
	mexPrintf( "   Examine input and output arguments in MLGAMain : Stop\n" );

	// Read arrays and sizes
	inXsize =    (int) UserData.ReadDouble(sXsize);
				       UserData.ReadArray(sXopt,&Xm,&Xn);
	inDsize =    (int) UserData.ReadDouble(sDsize);
				       UserData.ReadArray(sDwei,&Dm,&Dn);
	if (Xm*Xn!=inXsize) 
		mexPrintf ( " WARNING ! wrong size for Xval : Xm*Xn!=inXsize %d %d %d", Xm,Xn,inXsize );
	if (Dm*Dn!=inDsize) 
		mexPrintf ( " WARNING ! wrong size for Dwei : Dm*Dn!=inDsize %d %d %d", Dm,Dn,inDsize );

	mexPrintf( "******************************************************\n" );
	mexPrintf( " GO : Call of Calculate in MLGAMain, from Main\n" );
	mexPrintf( "******************************************************\n" );

	OptOBJfun = Calculate( SideFun, &UserData, &Parametres, &ControlData );

	mexPrintf( "******* Just after Calculate *********************************\n" );

	SideFunNum = SideFunDim;							// (following lines for zesting purposes only)
	for(i=0; i<SideFunDim; ++i) 
		mexPrintf ("$$$  SideFun[%i] = %f\n", i, SideFun[i]);	//	!?! to verify !

//	mexPrintf( "**************************************************************************\n" );
	mexPrintf( " Returning from Calculate, at the end of Main MEX start program !\n" );
/*	mexPrintf ( " Best genome values   inXval[] =\n" );
	mexPrintf( "**** Start UserData.PrintArg()  ******************************************\n" );
	UserData.PrintArg();
	mexPrintf( "**** End of UserData.PrintArg() ******************************************\n" );
*/
		
	// return User Data in [0]
	plhs[0] = UserData.GetArgs();

	// return Objective function in 1
	plhs[1]=mxCreateNumericArray(Taille,VirtualTable,mxDOUBLE_CLASS,mxREAL);
	memcpy(mxGetPr(plhs[1]), SideFun, VirtualTable[0]*sizeof(double));
	
	// return best value in 2
	plhs[2]=mxCreateDoubleMatrix(1,1,mxREAL);
	*mxGetPr(plhs[2]) = SideFunNum;		// output parameters !!!!!!!!!

}

///////////////////////////////////////////////
// Print info about parametres
///////////////////////////////////////////////
void MLGAMain::PrintParList( GAParameterList& ParList )
{
	int i;

	for( i = 0 ; i < ParList.size(); i++ )
		{
		mexPrintf( "%s (%s) =",ParList[i].fullname(), ParList[i].shrtname());
		switch( ParList[i].type())
			{
			case GAParameter::BOOLEAN:
				mexPrintf( "boolean %s",*(bool *)ParList[i].value()?"True":"False");
				break;
			case GAParameter::CHAR:
				mexPrintf( "char %c",*(char *)ParList[i].value());
				break;
			case GAParameter::STRING:
				mexPrintf( "string '%s'",ParList[i].value());
				break;
			case GAParameter::INT:
				mexPrintf( "int %d",*(int * )ParList[i].value());
				break;
			case GAParameter::FLOAT:
				mexPrintf( "float %f",*(float * )ParList[i].value());
				break;
			case GAParameter::DOUBLE:
				mexPrintf( "double %lf",*(double*)ParList[i].value());
				break;
			case GAParameter::POINTER:
				mexPrintf( "pointer %lx",ParList[i].value());
		} // Switch	

		mexPrintf( "\n" );
	} // For
	mexPrintf( "\n" );
} // PrintParList

///////////////////////////////////////////////////
//	To update selected parameters 
///////////////////////////////////////////////////
void MLGAMain::NewParam(GAParameterList *p,
						MLInterface *Parametres )
{
	int inN_POP = (int)Parametres->ReadDouble(sN_POP);
	int inN_GEN = (int)Parametres->ReadDouble(sN_GEN);
	float inP_MUT = Parametres->ReadDouble(sP_MUT);
	float inP_CRO = Parametres->ReadDouble(sP_CRO); 
	
	Parametres->PrintArg();

	p->add(gaNnGenerations, gaSNnGenerations, GAParameter::INT, &inN_GEN);
	p->add(gaNpCrossover, gaSNpCrossover, GAParameter::FLOAT, &inP_CRO);
	p->add(gaNpMutation, gaSNpMutation, GAParameter::FLOAT, &inP_MUT);
	p->add(gaNpopulationSize, gaSNpopulationSize, GAParameter::INT, &inN_POP);

	PrintParList( *p );

} // NewParam


///////////////////////////////////////////////////////////
// Generate population file
///////////////////////////////////////////////////////////
void MLGAMain::write_pop(ostream &fname,
			   StruggleHGA *ga,
			   int ArraySize) 
{
	fname << " Generation = " << ga->generation() << "\t" << "next line : chrom.score, then chrom.real_genome() & chrom.int_genome()[count](j=1," << ga->population().size() << ")\n";	
	for(int i=0; i<ga->population().size(); i++)
		{
		MixedGenome &chrom = (MixedGenome &)  ga->population().individual(i);
		MLEvalData *MLEV;
		MLEV = (MLEvalData *) ga->population().individual(i).evalData();
		fname  << " Ind=" << i+1 << "\tOF=" << chrom.score() <<
			"    \tHR_of_OF=" << MLEV->GetGenomeValue(ArraySize+5) << "    \t   with variables : " ;
		fname << "\n*real=";
		for(int jr = 0;jr<chrom.real_genome().length();++jr)
			fname << "  \t" << chrom.real_genome().gene(jr);
		fname << "\n* int=";
		for(int jn = 0;jn<chrom.int_genome().length();++jn)
			fname << "  \t" << chrom.int_genome().gene(jn);
		fname << "\n";
		}
	
	fname << " Current Objective function in current population = " << ga->statistics().current() << "\n";	
	fname << " Average Objective function in current population = " << ga->population().ave() << "\n";
	fname << " Best    Objective function in current population = " << ga->population().best().score() << "\n\n";
}

//////////////////////////////////////////
// to print out one genome
//////////////////////////////////////////
void MLGAMain::print_genome(ostream &ff,GAGenome& g,MLInterface *d) 
{
	MixedGenome& genome = (MixedGenome&)g;
	static double *x = NULL;
	int countreal,countint,i,m,n;
	double			Xsize = d->ReadDouble(sXsize);
	double			*inXopt = d->ReadArray(sXopt, &m, &n);
	double			*inXval = d->ReadArray(sXval, &m, &n);

	if(!x)
		x= new double[Xsize];

	if (MLIO::PrintTrace() > 0)	
		mexPrintf ( "==== Print complete genome x[1] to x[%d]\n",(int)Xsize);
	countreal=0;
	countint =0;
	for(i=0;i<Xsize;++i)
	{
		switch ((int)(10*inXopt[i]))
			{
			case 10 :
				x[i]=genome.int_genome().gene(countint);
				countint++;
				break;
			case 9 :
				x[i]=genome.real_genome().gene(countreal);
				countreal++;
				break;
			case 1 :
				x[i]=inXval[i];			
				break;
			case 0 :
				x[i]=inXval[i];			
				break;
			default :
				mexPrintf ( "Wrong value for Xopt (should be 0.0, 0.1, 0.9 or 1.0)\tinXopt[%d]=%lf\n\n\n ***STOP ! ***\n\n",i,inXopt[i] );
				mexPrintf("");
			}
		if (MLIO::PrintTrace() > 0)	
		{
			mexPrintf ( " %2i=%8.5f",i,x[i]);
			if (((i+1)/10)*10 == (i+1)) mexPrintf("\n");
		}
	} // for

	for(i=0;i<Xsize;++i) 
		{
		ff<<"GENOME :x["<<i<<"]="<<x[i]<<"\n";
		} // for

    ff << " The Objective function for this one best genome  = " << genome.score() << "\n";
}// print_genome


////////////////////////////////////////////////////////////
// Calculate function
////////////////////////////////////////////////////////////
double MLGAMain::Calculate(
					double *OF_SideFun,
					MLInterface *UserData,
					MLInterface *Parametres,
					MLInterface *ControlData)
{
	GAAlleleSetArray ra;
	GAAlleleSetArray na;
	int i;
	int CheckAve, DvarSize;
	double Average_TBC, Best_TBC, r_GEN;
	int PnPlot;
	int i_GEN, nGENtrace;


//	mexPrintf ("\n    MLGAMain::Calculate\n");
	
	// Teatement of Control data
	PnPlot = ControlData->ReadDouble(sPlot);
	if (PnPlot>0)
		{
		nGENtrace = (int)Parametres->ReadDouble(sN_GEN) / PnPlot;	//	plot frequency
		if (nGENtrace == 0) nGENtrace=1;							//	At least one plot
		}
	else 
		{
		PnPlot= (int)Parametres->ReadDouble(sN_GEN);
		nGENtrace=1;
		mexPrintf ("\n WARNING : PNPlot=0 ! Plot will be refreshed, each generation ! ");
		}

	ofstream BestIndOfGen;
	BestIndOfGen.open("BestIndOfGen.log", ios::trunc |ios::out);
	if(BestIndOfGen.bad()) 
		{
		std::cerr<<"Unable to open file 'BestIndOfGen.log'\n";
		exit(0);
		}

	ofstream population;
	population.open("population.txt", ios::trunc |ios::out);
	if(population.bad()) 
		{
		std::cerr<<"Unable to open file '"<<"pop.txt"<<"'\n";
		exit(0);
		}

	MLIO::PrintTrace( (int)ControlData->ReadDouble(sTrace) );
	mexPrintf(" MLGAMain::Calculate ----> Trace   = %i\n", MLIO::PrintTrace());
	MLIO::BCO_name( (char*)Parametres->ReadString(sBCOname) );
	mexPrintf(" MLGAMain::Calculate ----> BCOname = %s\n", MLIO::BCO_name());
	MLIO::BCO_Bmut( (double)Parametres->ReadDouble(sBCOBmut) );
	mexPrintf(" MLGAMain::Calculate ----> BCOBmut = %f\n", MLIO::BCO_Bmut());
	MLIO::BCO_Rdep( (double)Parametres->ReadDouble(sBCORdep) );
	mexPrintf(" MLGAMain::Calculate ----> BCORdep = %f\n", MLIO::BCO_Rdep());

	DvarSize = UserData->ReadDouble(sDsize);

	/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/
	// Output args creation
	int     n_out, n_in;
	mxArray *oa[1], *ia[4];
	n_out = 1;
	n_in  = 4;
	ia[0] = mxCreateDoubleMatrix(1, 1, mxREAL);
	ia[1] = mxCreateDoubleMatrix(1, 1, mxREAL);
	ia[2] = mxCreateDoubleMatrix(1, 1, mxREAL);
	ia[3] = mxCreateDoubleMatrix(1, 1, mxREAL);
	/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/

	// Adding genome data from arguments arrays 
	double nvals = UserData->ReadDouble(sXsize);
	for(i=0;i<nvals;++i) 
	{
		double *inXopt = UserData->ReadArray(sXopt);
		double *inXmin = UserData->ReadArray(sXmin);
		double *inXmax = UserData->ReadArray(sXmax);
		double *inXwei = UserData->ReadArray(sXwei);
		double *inXval = UserData->ReadArray(sXval);
		int incr_intallele = 1;
		switch ((int)(10*inXopt[i]))
		{
		case 10 :
			na.add(inXmin[i],inXmax[i],incr_intallele,GAAllele::INCLUSIVE,GAAllele::INCLUSIVE,inXwei[i]);
			BestIndOfGen <<"optimizing the x integer value \tinXval["<<i<<"] = "<<inXval[i]<<"\n";
			break;
		case 9 :
			ra.add(inXmin[i],inXmax[i],GAAllele::INCLUSIVE,GAAllele::INCLUSIVE,inXwei[i]);
			BestIndOfGen <<"optimizing the x real    value \tinXval["<<i<<"] = "<<inXval[i]<<"\n";
			break;
		case 1 :
			BestIndOfGen <<"fixed at initial real    value \tinXval["<<i<<"] = "<<inXval[i]<<"\n";
			break;
		case 0 :
			BestIndOfGen <<"fixed at initial integer value \tinXval["<<i<<"] = "<<inXval[i]<<"\n";
			break;
		default :
			BestIndOfGen <<"Wrong value for Xopt (should be 0.0, 0.1, 0.9 or 1.0)\t"<< i<< "\n\n\n ***STOP ! ***\n\n";
			mexPrintf ( "Wrong value for Xopt (should be 0.0, 0.1, 0.9 or 1.0)\tinXopt[%d]=%lf\n\n\n ***STOP ! ***\n\n",i,inXopt[i] );
			return(990099);
		}
	} // for
  
	BestIndOfGen<<"\n\nReal AlleleSetArray  \n\n"<<ra<<"\n============";
	BestIndOfGen<<"\n\nInt  AlleleSetArray  \n\n"<<na<<"\n============";
                                                     
	MixedGenome *mg =  new MixedGenome(ra,na,CalcObjective,(void *)UserData);

	/********** $$$$$$$   Start initialisation MLEvalData        $$$$$$$$ ************/
	/*	We are just after the basic definition and creation of our genome
		Now the goal is to add a MLEvalData object to Genome class */

	if (MLIO::PrintTrace() > 1)
		mexPrintf ( " Start                     ===> Greffe\n");

	MLEvalData EVDAmp;									// Original MLEvalData
	MLEvalData *EVDNew;									// A new pointer
	EVDNew = (MLEvalData *) (mg->evalData (EVDAmp));	// Clone EVDNew with EVDAmp

	if (MLIO::PrintTrace() > 1)
		mexPrintf( "Nb of clones left (Expected 2 !) = %i\n\n",MLEvalData::GetNbClone()); // Expected = 2

	if (MLIO::PrintTrace() > 1)	
		mexPrintf ("Address EVDNew & EVDAmp %d %d\n",EVDNew,EVDAmp);
	if (EVDNew == NULL) 
		mexPrintf ( " EVDNew == NULL   $$-$$\n");
	if (MLIO::PrintTrace() > 1)	
		mexPrintf ( " - Stick MLEvalData on genome in Calculate ! EVDNew->GetNbClone() = %d\n",EVDNew->GetNbClone() );
	if (MLIO::PrintTrace() > 1)	
		mexPrintf ( " Stop                      ===> Greffe\n\n");
	/********** $$$$$$$   End  initialisation MLEvalData         $$$$$$$ ************/

	mg->comparator(my_comparator);						// Definitions
	mg->crossover(MixedGenome::MixedCrossover);
//	mg->crossover(GARealBlendCrossover);

	if (MLIO::PrintTrace() > 1)	
	{
		mexPrintf( "Nb of clones created  = %i\n",MLEvalData::GetNbClone()); 
		mexPrintf ( " Before Struggle definition\n");
	}

	StruggleHGA *hga = new StruggleHGA((GAGenome &)*mg);
	if (MLIO::PrintTrace() > 1)	
	{
		mexPrintf ( " After  Struggle definition\n");
		mexPrintf( "Nb of clones left (Expected 2 ?) = %i\n",MLEvalData::GetNbClone());
	}

	// modifie les parametres
	GAParameterList *ParList;
	if (MLIO::PrintTrace() > 0)	
		mexPrintf ( "\n Parameter List\n");
  
	if (MLIO::PrintTrace() > 1)	
		{
		ParList = new GAParameterList;
		*ParList = hga->parameters();			// default parameter list
		PrintParList( *ParList );
		delete ParList;
		}

	hga->parameters(SETTINGS_FILE);
  
	if (MLIO::PrintTrace() > 1)	
		{
		ParList = new GAParameterList;
		*ParList = hga->parameters();			// new values read from file
		PrintParList( *ParList );
		delete ParList;
		}

	ParList = new GAParameterList;
	NewParam( ParList,Parametres);	
	// actualise les donnes des parametres
	hga->parameters(*ParList);
	if (MLIO::PrintTrace() > 0)
	{
		*ParList = hga->parameters();			// new values from passed arguments
		PrintParList( *ParList );
	}
	delete ParList;

	if (MLIO::PrintTrace() > 1)	
		mexPrintf( "Nb of clones left = %i\n",MLEvalData::GetNbClone()); 

	mexPrintf( "\n*****************************************************\n" );
	mexPrintf( "In Calculate (begin hga loop)" );
	mexPrintf( "\n*****************************************************\n" );
  

	mexPrintf("* hga->minimize();\n");
	hga->minimize();
	mexPrintf("* hga->initialize();\n");
	hga->initialize();

	mexPrintf("* Before while(!hga->done()) ?\n");
	while(!hga->done() )
		{
		if (MLIO::PrintTrace() > 0)	
			mexPrintf("In while(!hga->done()) : Take the best\n");
		MixedGenome & chrom = (MixedGenome &)(hga->population().best());
		i_GEN = hga->generation();
		if ( (i_GEN / nGENtrace)*nGENtrace == i_GEN )
				write_pop(population,hga,DvarSize);
		BestIndOfGen << " Generation no " << i_GEN << ", the best genome in this population is \n";
		print_genome ( BestIndOfGen, chrom, UserData );
		Average_TBC = hga->population().ave();
		Best_TBC    = hga->population().best().score();
		BestIndOfGen << " Average Objective function in current population = " << Average_TBC << "\n\n";
		if (MLIO::PrintTrace() > 0)	
			mexPrintf( "\nScore of the best genome    : hga->population().best().score() = %f\n", Best_TBC );
		if (MLIO::PrintTrace() > 0)	
			mexPrintf( "Average score of population : hga->population().ave() = %f\n", Average_TBC );
		/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/
		if (MLIO::PrintTrace() > 1)	
			mexPrintf ( " Just before mexCallMATLAB(n_out, oa, n_in, ia, ""RAve"" )\n" );
		MLEvalData *MLEV;
		MLEV = (MLEvalData *) hga->population().best().evalData();
		MLEV->SetGenomeValue(DvarSize+5,Average_TBC);
		*mxGetPr(ia[0]) = MLEV->GetGenomeValue ( DvarSize+4 );	//	Value_TBC
		*mxGetPr(ia[1]) = MLEV->GetGenomeValue ( DvarSize+5 );	//	Average_TBC
		*mxGetPr(ia[2]) = MLEV->GetGenomeValue ( DvarSize+6 );	//	HC_TBC
		*mxGetPr(ia[3]) = (double) i_GEN;
		if (MLIO::PrintTrace() > 1)	
		{
			mexPrintf ( " PRINT ! mxGetPr(ia[0]) =  %f\n", *mxGetPr(ia[0]) );
			mexPrintf ( " PRINT ! mxGetPr(ia[1]) =  %f\n", *mxGetPr(ia[1]) );
			mexPrintf ( " PRINT ! mxGetPr(ia[2]) =  %f\n", *mxGetPr(ia[2]) );
			mexPrintf ( " PRINT ! mxGetPr(ia[3]) =  %f\n", *mxGetPr(ia[3]) );
			mexPrintf ( "----------------\n" );
		}
		if (MLIO::PrintTrace() > 1)	
			mexPrintf( "Nb of clones before = %i\n", MLEV->GetNbClone() );

		mexCallMATLAB(n_out, oa, n_in, ia, "RAve" );
		
		if (MLIO::PrintTrace() > 1)	
			mexPrintf ( " Just after  mexCallMATLAB(n_out, oa, n_in, ia, ""RAve"" )\n\n" );

		if (MLIO::PrintTrace() > 1)	
			mexPrintf( "Nb of clones after = %i\n", MLEV->GetNbClone() );

		CheckAve = (int) mxGetScalar(oa[0]);
		if ((CheckAve != 988) && (CheckAve != 989))
			mexErrMsgTxt ( " ERROR in callto MATLAB RAve.m ! Check ! " );
		mxDestroyArray( oa[0] );

		/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/
		if ( (i_GEN / nGENtrace)*nGENtrace == i_GEN )
				hga->statistics().write(BestIndOfGen);

//		write_pop(population,hga,DvarSize);
		hga->step();
		if (MLIO::PrintTrace() > 1)	
			mexPrintf("Apres Step\n");
//		write_pop(population,hga,DvarSize);

	} // while

    MixedGenome & chrom = (MixedGenome &)(hga->population().best());
    write_pop(population,hga,DvarSize);
	i_GEN = hga->generation();
    BestIndOfGen << "\n ====>> Generation no " << i_GEN << ", the best genome in this population is \n";
	print_genome ( BestIndOfGen, chrom, UserData );
	Average_TBC = hga->population().ave();
	Best_TBC    = hga->population().best().score();
	BestIndOfGen << " Average Objective function in current population = " << Average_TBC << "\n\n";
	BestIndOfGen << " Statistics of the current population :\n";
	if (MLIO::PrintTrace() > 0)	
		mexPrintf( "\nScore of the best genome  : hga->population().best().score() = %f\n", Best_TBC );
	if (MLIO::PrintTrace() > 0)	
		mexPrintf( "Average score of population : hga->population().ave() = %f\n", Average_TBC );
	/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/
	if (MLIO::PrintTrace() > 1)
		mexPrintf ( " Just before mexCallMATLAB( n_out, oa, n_in, ia, ""RAve"" )\n" );

	MLEvalData *MLEV;
	MLEV = (MLEvalData *) (hga->population().best().evalData());
		MLEV->SetGenomeValue(DvarSize+5,Average_TBC);
		int OF_SideFunDim = (int)(MLEV->GetGenomeValue ( DvarSize ));	//	Number of functions
		*mxGetPr(ia[0]) = MLEV->GetGenomeValue ( DvarSize+4 );		//	Value_TBC
		*mxGetPr(ia[1]) = MLEV->GetGenomeValue ( DvarSize+5 );		//	Average_TBC
		*mxGetPr(ia[2]) = MLEV->GetGenomeValue ( DvarSize+6 );		//	HC_TBC
		*mxGetPr(ia[3]) = (double) i_GEN;		// Normal treatment in Rave for all call, except last
	if (MLIO::PrintTrace() > 1)
	{
		mexPrintf ( " PRINT ! mxGetPr(ia[0]) =  %f\n", *mxGetPr(ia[0]) );
		mexPrintf ( " PRINT ! mxGetPr(ia[1]) =  %f\n", *mxGetPr(ia[1]) );
		mexPrintf ( " PRINT ! mxGetPr(ia[2]) =  %f\n", *mxGetPr(ia[2]) );
		mexPrintf ( " PRINT ! mxGetPr(ia[3]) =  %f\n", *mxGetPr(ia[3]) );
		mexPrintf ( " \n=====================\n");
	}

	for(i=0; i<OF_SideFunDim; ++i)
		OF_SideFun[i] = MLEV->GetGenomeValue ( DvarSize+i+1 );				//	OK !
		mexPrintf ("$$$$$$$$$$$$$  OF_SideFunDim = %i\n", OF_SideFunDim);		//	OK !
	for(i=0; i<OF_SideFunDim; ++i)
		mexPrintf ("$$$$$$$$$$$$$  OF_SideFun[%i] = %f\n", i, OF_SideFun[i]);		//	OK !
	
	mexCallMATLAB(n_out, oa, n_in, ia, "RAve" );

	if (MLIO::PrintTrace() > 1)	
		mexPrintf ( " Just after  mexCallMATLAB(n_out, oa, n_in, ia, ""RAve"" )\n\n" );

	CheckAve = (int) mxGetScalar(oa[0]);
	if (CheckAve != 989) 
		mexErrMsgTxt ( " ERROR in callto MATLAB RAve.m ! Check ! " );
	mxDestroyArray( oa[0] );
	mxDestroyArray( ia[0] );	// At the final end only !
	mxDestroyArray( ia[1] );
	mxDestroyArray( ia[2] );
	mxDestroyArray( ia[3] );
	/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/
	hga->statistics().write(BestIndOfGen);

	//	GAStatistics Mean;
	if (MLIO::PrintTrace() > 1)	
		mexPrintf( "In Calculate End : hga->statistics().current(GAStatistics::Mean) = %f\n",
									hga->statistics().current(GAStatistics::Mean) );

	population.close();
	BestIndOfGen.close();

	mexPrintf( "\n\nEnd1 - Nb of clones left = over %i\n\n",MLEvalData::GetNbClone());

	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	//	Return the best genome in inXval input array
	int countreal,countint;
	double MaxCount;
	double *OptArray;
	double *ValArray;

	MaxCount = UserData->ReadDouble(sXsize);	//	Total number of gene
	OptArray = UserData->ReadArray(sXopt);		//	optimization parameters
	ValArray = UserData->ReadArray(sXval);		//	initial values (real, integer)
	
	mexPrintf("MaxCount = %d\n", MaxCount );
	for(i=0,countreal=0,countint=0; i<(int)MaxCount; ++i) 
	{
		if		(OptArray[i] == 1.0)			// if this gene is in the integer genome
		{
			ValArray[i]=chrom.int_genome().gene(countint++);
			mexPrintf("1=* OptArray[%2d] = %f  --  ValArray[%2d] = %f\n", i, OptArray[i], i, ValArray[i] );
		}
		else if	(OptArray[i] == 0.9)			// if this gene is in the real genome
		{
			ValArray[i]=chrom.real_genome().gene(countreal++);
			mexPrintf("2=* OptArray[%2d] = %f  --  ValArray[%2d] = %f\n", i, OptArray[i], i, ValArray[i] );
		}
		else if	(OptArray[i] == 0.1)			// if this gene is in the real genome
			mexPrintf("3=* OptArray[%2d] = %f\n", i, OptArray[i] );
		else if	(OptArray[i] == 0.0)			// if this gene is in the integer genome
			mexPrintf("4=* OptArray[%2d] = %f\n", i, OptArray[i] );
		else									// if not
			mexPrintf ( "Wrong value for Xopt (should be 0.0, 0.1, 0.9 or 1.0)\tinXopt[%d]=%lf\n\n\n ***STOP ! ***\n\n",i,OptArray[i] );
	}
//	mexPrintf("--------------------------\n");
	/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
	// Clean up
	delete hga;
//	mexPrintf( "End2 - Nb of clones left = over %i\n",MLEvalData::GetNbClone());
	delete mg;									//	see if nvarInt = 0 ?  ==> erreur !
//	mexPrintf( "End3 - Nb of clones left = over %i\n",MLEvalData::GetNbClone());

	return(Best_TBC);
} // Calculate
 
