/* ----------------------------------------------------------------------------
  mixed.cpp
  akm 1April97
  Copyright (c) 1997 Adam K Molyneaux
                     all rights reserved

 DESCRIPTION:
	A mixed integer and real genome.

 MODIFICATIONS:
 21December2000 - Blaise Renaud
	- Created GAIntBlendCrossover, a direct translation from GARealBlendCrossover
		Enlarging by deploying the exploitation interval of each gene (100% both side)
		And taking care of the limits.
	- Created GABinBlendCrossover for binary genome
		Standard binary generation, with some kind of mutation if parents' genes are the same.
		
---------------------------------------------------------------------------- */
#include "ga/MixedGenome.h"
#include "mlio.h"
#include <stdio.h>
#include <string.h>


// Binary blend crossover generates a new value based on the interval between parents.
// We generate a uniform distribution based on the distance between parent
// values, then choose the child value based upon that distribution.
// taking care that it always obey limits on alleles.
// Added by Blaise 19.12.2000
int 
GABinBlendCrossover(const GAGenome& p1, const GAGenome& p2,
		     GAGenome* c1, GAGenome* c2) {
  GA1DArrayAlleleGenome &mom=(GA1DArrayAlleleGenome &)p1;
  GA1DArrayAlleleGenome &dad=(GA1DArrayAlleleGenome &)p2;

  int   n=0;
  float fmut = MLIO::BCO_Bmut();
  float fsis, fbro;
  int   isis, ibro;
  int	imomgene, idadgene;

	static bool isOpen;
	static ofstream Stream_File_MG_Bin;
	if (! isOpen)	{
		Stream_File_MG_Bin.open("MG_GABlendCrossover_Bin.log", ios::trunc |ios::out);
		isOpen = true;
	}
	if (MLIO::PrintTrace() > 0)	
	{
		Stream_File_MG_Bin<<"\n====  Start(GABinBlendCrossover)   with fmut = MLIO::BCO_Bmut() = "<<fmut;
		Stream_File_MG_Bin<<"\n=  mom.write(Stream_File_MG_Bin) ";
		mom.write(Stream_File_MG_Bin);
		Stream_File_MG_Bin<<"\n==== mom.alleleset("<<1<<").size()="<<mom.alleleset(0).size()<<"  mom.alleleset(0).lower()="<<mom.alleleset(0).lower()<<"  mom.alleleset(0).upper()="<<mom.alleleset(0).upper();
		Stream_File_MG_Bin<<"\n=  dad.write(Stream_File_MG_Bin) ";
		dad.write(Stream_File_MG_Bin);
		Stream_File_MG_Bin<<"\n==== dad.alleleset("<<1<<").size()="<<dad.alleleset(0).size()<<"  dad.alleleset(0).lower()="<<dad.alleleset(0).lower()<<"  dad.alleleset(0).upper()="<<dad.alleleset(0).upper();
	}

  if(c1 && c2){
    GA1DArrayAlleleGenome  &sis=(GA1DArrayAlleleGenome  &)*c1;
    GA1DArrayAlleleGenome  &bro=(GA1DArrayAlleleGenome  &)*c2;

    int len = GAMax(mom.length(), dad.length());
    for(int i=0; i<len; i++) {

	  imomgene = (int) (mom.gene(i) + 0.5);
	  idadgene = (int) (dad.gene(i) + 0.5);
	  if (MLIO::PrintTrace() > 1)
		Stream_File_MG_Bin<<"\n==* n=2 *== i="<<i<<"  imomgene="<<imomgene<<"  idadgene="<<idadgene;

      if(imomgene == idadgene)
	  {
		fsis = GARandomFloat(0,1);
		fbro = GARandomFloat(0,1);
		if(imomgene == 0)
		{
			if (fsis > fmut)	isis = 0;
			else				isis = 1;
			if (fbro > fmut)	ibro = 0;
			else				ibro = 1;
		}
		else 
		{
			if (fsis > fmut)	isis = 1;
			else				isis = 0;
			if (fbro > fmut)	ibro = 1;
			else				ibro = 0;
		}
		if (MLIO::PrintTrace() > 1)	
		    Stream_File_MG_Bin<<"\n==* n=2 *== i="<<i<<"  isis="<<isis<<"  ibro="<<ibro<<"  fsis="<<fsis<<"  fbro="<<fbro<<"  fmut="<<fmut;
		sis.gene(i, isis);
		bro.gene(i, ibro);
		
	  }
	  else
	  {
		sis.gene(i, GARandomInt(0, 1));
		bro.gene(i, GARandomInt(0, 1));
	  }

	}
	if (MLIO::PrintTrace() > 1)	
	{
		Stream_File_MG_Bin<<"\n=  sis->write(Stream_File_MG_Bin) ";
		sis.write(Stream_File_MG_Bin);
		Stream_File_MG_Bin<<"\n=  bro->write(Stream_File_MG_Bin) ";
		bro.write(Stream_File_MG_Bin);
	}
    n = 2;
  }
  else if(c1 || c2){
    GA1DArrayAlleleGenome &sis = 
      (c1 ? (GA1DArrayAlleleGenome &)*c1 : (GA1DArrayAlleleGenome &)*c2);

    int len = GAMax(mom.length(), dad.length());
    for(int i=0; i<len; i++) {

	  imomgene = (int) (mom.gene(i) + 0.5);
	  idadgene = (int) (dad.gene(i) + 0.5);
	  if (MLIO::PrintTrace() > 1)	
		Stream_File_MG_Bin<<"\n==* n=1 *== i="<<i<<"  imomgene="<<imomgene<<"  idadgene="<<idadgene;

      if(imomgene == idadgene)
	  {
		fsis = GARandomFloat(0,1);
		if(imomgene == 0)
		{
			if (fsis > fmut)	isis = 0;
			else				isis = 1;
		}
		else 
		{
			if (fsis > fmut)	isis = 1;
			else				isis = 0;
		}
		if (MLIO::PrintTrace() > 1)	
		    Stream_File_MG_Bin<<"\n==* n=1 *== i="<<i<<"  isis="<<isis<<"  fsis="<<fsis<<"  fmut="<<fmut;
		sis.gene(i, isis);
		
	  }
	  else
	  {
		sis.gene(i, GARandomInt(0, 1));
	  }

	}
	if (MLIO::PrintTrace() > 1)	
	{
		Stream_File_MG_Bin<<"\n=  sis->write(Stream_File_MG_Bin) ";
		sis.write(Stream_File_MG_Bin);
	}
    n = 1;
  }
  if (isOpen)
  {
	Stream_File_MG_Bin.close();
	isOpen = false;
  }
  return n;
}


// Integer blend crossover generates a new value based on the interval between parents.
// We generate a uniform distribution based on the distance between parent
// values, then choose the child value based upon that distribution.
// taking care that it always obey limits on alleles.
// Added by Blaise 15.12.2000 (inspired from GARealBlendCrossover, with deployment = 1 rather than 0.5)
int 
GAIntBlendCrossover(const GAGenome& p1, const GAGenome& p2,
		     GAGenome* c1, GAGenome* c2) {
  GA1DArrayAlleleGenome &mom=(GA1DArrayAlleleGenome &)p1;
  GA1DArrayAlleleGenome &dad=(GA1DArrayAlleleGenome &)p2;

  int n=0;

	static bool isOpen;
	static ofstream Stream_File_MG_Int;
	if (! isOpen)	{
		Stream_File_MG_Int.open("MG_GABlendCrossover_Int.log", ios::trunc |ios::out);
		isOpen = true;
	}
	if (MLIO::PrintTrace() > 0)	
	{
		Stream_File_MG_Int<<"\n====  Start(GAIntBlendCrossover)";
		Stream_File_MG_Int<<"\n=  mom.write(Stream_File_MG_Int) ";
		mom.write(Stream_File_MG_Int);
		Stream_File_MG_Int<<"\n==== mom.alleleset(0).size()="<<mom.alleleset(0).size()<<"  mom.alleleset(0).lower()="<<mom.alleleset(0).lower()<<"  mom.alleleset(0).upper()="<<mom.alleleset(0).upper();
		Stream_File_MG_Int<<"\n=  dad.write(Stream_File_MG_Int) ";
		dad.write(Stream_File_MG_Int);
		Stream_File_MG_Int<<"\n==== dad.alleleset(0).size()="<<dad.alleleset(0).size()<<"  dad.alleleset(0).lower()="<<dad.alleleset(0).lower()<<"  dad.alleleset(0).upper()="<<dad.alleleset(0).upper();
	}

  if(c1 && c2){
    GA1DArrayAlleleGenome  &sis=(GA1DArrayAlleleGenome  &)*c1;
    GA1DArrayAlleleGenome  &bro=(GA1DArrayAlleleGenome  &)*c2;

    int len = GAMax(mom.length(), dad.length());
    for(int i=0; i<len; i++) {
      float dist = 0;
      if(mom.gene(i) > dad.gene(i)) 
		dist = mom.gene(i) - dad.gene(i);
      else 
		dist = dad.gene(i) - mom.gene(i);
	  int idist = (int)(dist+0.5);
	if (MLIO::PrintTrace() > 1)	
		Stream_File_MG_Int<<"\n==* n=2 *== i="<<i<<"  dist="<<dist<<"  idist="<<idist<<"  mom.gene("<<i<<")="<<mom.gene(i)<<"  dad.gene("<<i<<")="<<dad.gene(i);

      int lo = GAMax(((int)(GAMin(mom.gene(i), dad.gene(i))+0.5) - idist),(int)mom.alleleset(i).lower());
      int hi = GAMin(((int)(GAMax(mom.gene(i), dad.gene(i))+0.5) + idist),(int)mom.alleleset(i).upper());
	if (MLIO::PrintTrace() > 1)	
	  Stream_File_MG_Int<<"\n=== n=2 === (0.5*dist)="<<(0.5*dist)<<"    (idist)="<<(idist)<<" === lo="<<lo<<"    hi="<<hi;
      sis.gene(i, GARandomInt(lo, hi));
      bro.gene(i, GARandomInt(lo, hi));
	}
	if (MLIO::PrintTrace() > 1)	
	{
		Stream_File_MG_Int<<"\n=  sis->write(Stream_File_MG_Int) ";
		sis.write(Stream_File_MG_Int);
		Stream_File_MG_Int<<"\n=  bro->write(Stream_File_MG_Int) ";
		bro.write(Stream_File_MG_Int);
	}
    n = 2;
  }
  else if(c1 || c2){
    GA1DArrayAlleleGenome &sis = 
      (c1 ? (GA1DArrayAlleleGenome &)*c1 : (GA1DArrayAlleleGenome &)*c2);

    int len = GAMax(mom.length(), dad.length());
    for(int i=0; i<len; i++) {
      float dist = 0;
      if(mom.gene(i) > dad.gene(i)) 
		dist = mom.gene(i) - dad.gene(i);
      else 
		dist = dad.gene(i) - mom.gene(i);
	  int idist = (int)(dist+0.5);
	  if (MLIO::PrintTrace() > 1)	
		Stream_File_MG_Int<<"\n==* n=2 *== i="<<i<<"  dist="<<dist<<"  idist="<<idist<<"  mom.gene("<<i<<")="<<mom.gene(i)<<"  dad.gene("<<i<<")="<<dad.gene(i);

      int lo = GAMax(((int)(GAMin(mom.gene(i), dad.gene(i))+0.5) - idist),(int)mom.alleleset(i).lower());
      int hi = GAMin(((int)(GAMax(mom.gene(i), dad.gene(i))+0.5) + idist),(int)mom.alleleset(i).upper());
	  if (MLIO::PrintTrace() > 1)	
		Stream_File_MG_Int<<"\n=== n=2 === (0.5*dist)="<<(0.5*dist)<<"    (int)(0.5*dist)="<<(int)(0.5*dist)<<" === lo="<<lo<<"    hi="<<hi;
      sis.gene(i, GARandomInt(lo, hi));
	}
	if (MLIO::PrintTrace() > 1)	
	{
		Stream_File_MG_Int<<"\n=  sis->write(Stream_File_MG_Int) ";
		sis.write(Stream_File_MG_Int);
	}
    n = 1;
  }
  if (MLIO::PrintTrace() > 0)	
	Stream_File_MG_Int<<"\n====  End\n";
  if (isOpen)
  {
	Stream_File_MG_Int.close();
	isOpen = false;
  }
  return n;
}


// Real blend crossover generates a new value based on the interval between parents.
// We generate a uniform distribution based on the distance between parent
// values, then choose the child value based upon that distribution.
// modified Adam 23Nov96 to always obey limits
// on alleles.
int 
GARealBlendCrossover(const GAGenome& p1, const GAGenome& p2,
		     GAGenome* c1, GAGenome* c2) {
  GA1DArrayAlleleGenome &mom=(GA1DArrayAlleleGenome &)p1;
  GA1DArrayAlleleGenome &dad=(GA1DArrayAlleleGenome &)p2;

  int n=0;
  float deploy = MLIO::BCO_Rdep();

	static bool isOpen;
	static ofstream Stream_File_MG_Real;
	if (! isOpen)	{
		Stream_File_MG_Real.open("MG_GABlendCrossover_Real.log", ios::trunc |ios::out);
		isOpen = true;
	}
	if (MLIO::PrintTrace() > 0)						// 2 only for real case
	{
		Stream_File_MG_Real<<"\n====  Start(GARealBlendCrossover)   with deploy = MLIO::BCO_Rdep() = "<<deploy;
		Stream_File_MG_Real<<"\n=  mom.write(Stream_File_MG_Real) ";
		mom.write(Stream_File_MG_Real);
		Stream_File_MG_Real<<"\n=  dad.write(Stream_File_MG_Real) ";
		dad.write(Stream_File_MG_Real);
	}

  if(c1 && c2){
    GA1DArrayAlleleGenome  &sis=(GA1DArrayAlleleGenome  &)*c1;
    GA1DArrayAlleleGenome  &bro=(GA1DArrayAlleleGenome  &)*c2;

    int len = GAMax(mom.length(), dad.length());
    for(int i=0; i<len; i++) {
      float dist = 0;
      if(mom.gene(i) > dad.gene(i)) 
		dist = mom.gene(i) - dad.gene(i);
      else 
		dist = dad.gene(i) - mom.gene(i);
	  if (MLIO::PrintTrace() > 2)	
		Stream_File_MG_Real<<"\n==* n=2 *== i="<<i<<"  dist="<<dist<<"  mom.gene("<<i<<")="<<mom.gene(i)<<"  dad.gene("<<i<<")="<<dad.gene(i);
      float lo = GAMax(((float)(GAMin(mom.gene(i), dad.gene(i))) - deploy*dist),(float)mom.alleleset(i).lower());
      float hi = GAMin(((float)(GAMax(mom.gene(i), dad.gene(i))) + deploy*dist),(float)mom.alleleset(i).upper());
	  if (MLIO::PrintTrace() > 2)	
		Stream_File_MG_Real<<"\n=== n=2 === (deploy*dist)="<<(deploy*dist)<<" === lo="<<lo<<"    hi="<<hi;
      sis.gene(i, GARandomFloat(lo, hi));
      bro.gene(i, GARandomFloat(lo, hi));
    }
	if (MLIO::PrintTrace() > 1)	
	{
		Stream_File_MG_Real<<"\n=  sis->write(Stream_File_MG_Real) ";
		sis.write(Stream_File_MG_Real);
		Stream_File_MG_Real<<"\n=  bro->write(Stream_File_MG_Real) ";
		bro.write(Stream_File_MG_Real);
	}
    n = 2;
  }
  else if(c1 || c2){
    GA1DArrayAlleleGenome &sis = 
      (c1 ? (GA1DArrayAlleleGenome &)*c1 : (GA1DArrayAlleleGenome &)*c2);

    int len = GAMax(mom.length(), dad.length());
    for(int i=0; i<len; i++) {
      float dist = 0;
      if(mom.gene(i) > dad.gene(i)) 
		dist = mom.gene(i) - dad.gene(i);
      else 
		dist = dad.gene(i) - mom.gene(i);
	  if (MLIO::PrintTrace() > 2)	
		Stream_File_MG_Real<<"\n==* n=2 *== i="<<i<<"  dist="<<dist<<"  mom.gene("<<i<<")="<<mom.gene(i)<<"  dad.gene("<<i<<")="<<dad.gene(i);
      float lo = GAMax(((float)(GAMin(mom.gene(i), dad.gene(i))) - deploy*dist),(float)mom.alleleset(i).lower());
      float hi = GAMin(((float)(GAMax(mom.gene(i), dad.gene(i))) + deploy*dist),(float)mom.alleleset(i).upper());
	  if (MLIO::PrintTrace() > 2)	
		Stream_File_MG_Real<<"\n=== n=2 === (deploy*dist)="<<(deploy*dist)<<" === lo="<<lo<<"    hi="<<hi;
      sis.gene(i, GARandomFloat(lo, hi));
    }
	if (MLIO::PrintTrace() > 1)	
	{
		Stream_File_MG_Real<<"\n=  sis->write(Stream_File_MG_Real) ";
		sis.write(Stream_File_MG_Real);
	}
    n = 1;
  }
  if (MLIO::PrintTrace() > 0)	
	Stream_File_MG_Real<<"\n====  End\n";
  if (isOpen)
  {
	Stream_File_MG_Real.close();
	isOpen = false;
  }
  return n;
}


// List blend crossover generates a new value based on the interval between parents.
// We generate a uniform distribution based on the distance between parent
// values, then choose the child value based upon that distribution.
// modified Adam 23Nov96 to always obey limits
// on alleles.

int GAListBlendCrossover(const GAGenome& p1, const GAGenome& p2,
		     GAGenome* c1, GAGenome* c2) {
  GA1DArrayAlleleGenome &mom=(GA1DArrayAlleleGenome &)p1;
  GA1DArrayAlleleGenome &dad=(GA1DArrayAlleleGenome &)p2;
  register int i,j;
  int n=0;

	static bool isOpen;
	static ofstream Stream_File_MG_List;
	if (! isOpen)	{
		Stream_File_MG_List.open("MG_GABlendCrossover_List.log", ios::trunc |ios::out);
		isOpen = true;
	}
	if (MLIO::PrintTrace() > 0)
	{
		Stream_File_MG_List<<"\n====  Start(GAListBlendCrossover)";
		Stream_File_MG_List<<"\n=  mom.write(Stream_File_MG_List) ";
		mom.write(Stream_File_MG_List);
		Stream_File_MG_List<<"\n=  dad.write(Stream_File_MG_List) ";
		dad.write(Stream_File_MG_List);
	}

  if(c1 && c2){
    GA1DArrayAlleleGenome &sis=(GA1DArrayAlleleGenome &)*c1;
    GA1DArrayAlleleGenome &bro=(GA1DArrayAlleleGenome &)*c2;

    int len = mom.length();	// assumes both same length
    int dist,momindx,dadindx,found;
    for(i=0; i<len; i++) {	// for each gene
      found=0;
      dist=0;
      for(j=0;j<mom.alleleset(i).size()-1;j++) {
	if(mom.gene(i) == mom.alleleset(i).allele(j)) {
	  if (MLIO::PrintTrace() > 1)	
		Stream_File_MG_List<<"\n==* n=2 *== j="<<j<<"  mom.alleleset("<<i<<").size()="<<mom.alleleset(i).size()<<"  mom.alleleset(i).allele(j)="<<mom.alleleset(i).allele(j);
	  momindx = j;
	  found++;
	  if(found == 2)
	    break;
	}
	if(dad.gene(i) == dad.alleleset(i).allele(j)) {
	  if (MLIO::PrintTrace() > 1)	
		Stream_File_MG_List<<"\n==* n=2 *== j="<<j<<"  dad.alleleset("<<i<<").size()="<<dad.alleleset(i).size()<<"  dad.alleleset(i).allele(j)="<<dad.alleleset(i).allele(j);
	  dadindx = j;
	  found++;
	  if(found == 2)
	    break;
	}
      }
				// now have the two indx for the parents
      if(dadindx > momindx) 
	dist = dadindx-momindx;
      else
	dist = momindx-dadindx;
      
      if(dist > 1) {
	dist =  dist/2;
      }
      
      int lo = GAMax((int)0,(int)GAMin(dadindx,momindx)-dist);
      int hi = GAMin(mom.alleleset(i).size()-1,(int)GAMin(dadindx,momindx)+dist);

	  if (MLIO::PrintTrace() > 1)	
		Stream_File_MG_List<<"\n==* n=2 *==  len="<<len<<",dadindx="<<dadindx<<",momindx="<<momindx<<",dist="<<dist<<",lo="<<lo<<",hi="<<hi;

      sis.gene(i,mom.alleleset(i).allele(GARandomInt(lo,hi)));
      bro.gene(i,mom.alleleset(i).allele(GARandomInt(lo,hi)));
    }
    n = 2;
	if (MLIO::PrintTrace() > 1)	
	{
		Stream_File_MG_List<<"\n=  bro->write(Stream_File_MG_List) ";
		bro.write(Stream_File_MG_List);
		Stream_File_MG_List<<"\n=  sis->write(Stream_File_MG_List) ";
		sis.write(Stream_File_MG_List);
	}
  }
  else if(c1 || c2){
    GA1DArrayAlleleGenome &sis=
      (c1 ? (GA1DArrayAlleleGenome &)*c1:(GA1DArrayAlleleGenome &)*c2);

    int len = mom.length();	// assumes both same length
    int dist,momindx,dadindx,found;
    for( i=0; i<len; i++) {	// for each gene
      found=0;
      dist=0;
      for(j=0;j<mom.alleleset(i).size()-1;j++) {
	if(mom.gene(i) == mom.alleleset(i).allele(j)) {
	  if (MLIO::PrintTrace() > 1)	
		Stream_File_MG_List<<"\n==* n=1 *== j="<<j<<"  mom.alleleset("<<i<<").size()="<<mom.alleleset(i).size()<<"  mom.alleleset(i).allele(j)="<<mom.alleleset(i).allele(j);
	  momindx = j;
	  found++;
	  if(found >1)
	    break;
	}
	if(dad.gene(i) == dad.alleleset(i).allele(j)) {
	  if (MLIO::PrintTrace() > 1)	
		Stream_File_MG_List<<"\n==* n=1 *== j="<<j<<"  dad.alleleset("<<i<<").size()="<<dad.alleleset(i).size()<<"  dad.alleleset(i).allele(j)="<<dad.alleleset(i).allele(j);
	  dadindx = j;
	  found++;
	  if(found >1)
	    break;
	}
      }
				// now have the two indx for the parents
      if(dadindx > momindx) 
	dist = dadindx-momindx;
      else
	dist = momindx-dadindx;
      if(dist > 1) {
	dist =  dist/2;
      }
      int lo = GAMax((int)0,(int)GAMin(dadindx,momindx)-dist);
      int hi = GAMin(mom.alleleset(i).size()-1,(int)GAMin(dadindx,momindx)+dist);

	  if (MLIO::PrintTrace() > 1)	
		Stream_File_MG_List<<"\n==* n=1 *==  len="<<len<<",dadindx="<<dadindx<<",momindx="<<momindx<<",dist="<<dist<<",lo="<<lo<<",hi="<<hi;

      sis.gene(i,mom.alleleset(i).allele(GARandomInt(lo,hi)));
    }
    n = 1;
	if (MLIO::PrintTrace() > 1)
	{
		Stream_File_MG_List<<"\n=  sis->write(Stream_File_MG_List) ";
		sis.write(Stream_File_MG_List);
	}
  }
  if (MLIO::PrintTrace() > 0)	
	Stream_File_MG_List<<"\n====  End\n";
  if (isOpen)
  {
	Stream_File_MG_List.close();
	isOpen = false;
  }
  return n;
}


// Member functions for the composite genome object
MixedGenome::MixedGenome(GAAlleleSetArray &float_allele,GAAlleleSetArray &int_allele, GAGenome::Evaluator f, void* u):
GAGenome(MixedInitializer,
	 MixedMutator,
	 MixedComparator) 
{
    evaluator(f); 
    userData(u); 
    crossover(MixedCrossover);
    _greal = new GA1DArrayAlleleGenome (float_allele,  f, u);
    _greal->crossover(GARealBlendCrossover);
    _gint = new GA1DArrayAlleleGenome (int_allele,  f, u);

	if      (_stricmp(MLIO::BCO_name(),"Bin") == 0)								// Kc27
	{
		_gint->crossover(GABinBlendCrossover);							// Kc19
		printf ( "++ MixedGenome::MixedGenome ++ GABinBlendCrossover\n" );
	}
	else if (_stricmp(MLIO::BCO_name(),"Int") == 0)
	{
		_gint->crossover(GAIntBlendCrossover);							// Kc18
		printf ( "++ MixedGenome::MixedGenome ++ GAIntBlendCrossover\n" );
	}
	else if (_stricmp(MLIO::BCO_name(),"List") == 0)
	{
		_gint->crossover(GAListBlendCrossover);							// Kc16
		printf ( "++ MixedGenome::MixedGenome ++ GAListBlendCrossover\n" );
	}
	else if (_stricmp(MLIO::BCO_name(),"Real") == 0)
	{
		_gint->crossover(GARealBlendCrossover);							// Kc14
		printf ( "++ MixedGenome::MixedGenome ++ GARealBlendCrossover\n" );
	}
	else
	{
		printf ( "++ MixedGenome::MixedGenome ++ Wrong BlendCrossover name value (Bin Int List Real)   MLIO::BCO_name()=%s+++\n *** STOP ! ***\n\n",
						MLIO::BCO_name() );
	}
	printf(" End of MixedGenome::MixedGenome ************************\n");
}

MixedGenome::MixedGenome(const MixedGenome & orig) {
  _greal = new GA1DArrayAlleleGenome (orig.real_genome());
  _gint = new GA1DArrayAlleleGenome (orig.int_genome());
  copy(orig);
}
  
MixedGenome& MixedGenome::operator=(const GAGenome& g) { copy(g); return *this; }

MixedGenome::~MixedGenome() { delete _greal; delete _gint; }

GAGenome* MixedGenome::clone(GAGenome::CloneMethod) const {
  return new MixedGenome(*this);
}

void MixedGenome::copy(const GAGenome & c){
  if(&c != this && sameClass(c)){
    GAGenome::copy(c);
    MixedGenome & bc = (MixedGenome &)c;
    _greal->copy(*(bc._greal));
    _gint->copy(*(bc._gint));
  }
}

int MixedGenome::equal(const GAGenome& g) const {
  MixedGenome& genome = (MixedGenome&)g;
  GA1DArrayAlleleGenome &g1 = (GA1DArrayAlleleGenome &) (this->real_genome());
  GA1DArrayAlleleGenome &g2 = (GA1DArrayAlleleGenome &) (genome.real_genome());
  GA1DArrayAlleleGenome &h1 = (GA1DArrayAlleleGenome &) (this->int_genome());
  GA1DArrayAlleleGenome &h2 = (GA1DArrayAlleleGenome &) (genome.int_genome());
  return(g1.equal(g2) && h1.equal(h2));
//  return (((*_greal) == (*(genome._greal))) && ((*_gint) == (*(genome._gint))));
}

int MixedGenome::read(istream & is) {
  float sc=0.0;

  is>>*_greal>>*_gint;
  is>>sc;
  this->score(sc);// set the score
  return is.fail() ? 1 : 0; 
}

int MixedGenome::write(ostream & os) const {
  
  os << "MixedGenome::write(ostream & os) = " << os << "\n";
  os << "*_greal = " << *_greal << "\n*_gint = " << *_gint;
  os << "\n_score = " << _score << "\n";

  return os.fail() ? 1 : 0;
}



// These are the default initialization, mutation, and comparator operators for
// this genome class.  They are defined as static functions of the composite
// genome class and they're defaults for the class.  But they can be overridden
// on any instance of the genome.

// The initializer just calls the initializer for each of the genomes that are
// in the composite genome.

// I would have used simply 'Initializer', 'Mutator', etc rather than
// 'CompositeInitializer' but old versions of g++ are brain-dead and don't
// get the encapsulation properly.
void MixedGenome::MixedInitializer(GAGenome & c) {
  MixedGenome & child = (MixedGenome &)c;
  child.real_genome().initialize();
  cerr << "OK\n";
  child.int_genome().initialize();
}


// The mutator just calls the mutator for each of the component genomes.
int MixedGenome::MixedMutator(GAGenome & c, float pmut) {
  MixedGenome & child = (MixedGenome &)c;
  return child.real_genome().mutate(pmut) + child.int_genome().mutate(pmut);
}

// The comparator just calls the comparators for each of the component genomes,
// then averages the score.
float MixedGenome::MixedComparator(const GAGenome& a, const GAGenome& b) {
  fprintf(stderr,"called MixedComparator - NOT WRITTEN YET !!!!\n");
  exit(2);
  MixedGenome& sis = (MixedGenome &)a;
  MixedGenome& bro = (MixedGenome &)b;
  //  return 0.5 * (sis.binstr().compare(bro) + sis.bin2dec().compare(bro));
  return(0);
}

// The crossover operator invokes the crossover for each of the genomes in the
// composite genome.  We use sexual crossover only, and we do not test to see
// if no crossover has been assigned.
int
MixedGenome::
MixedCrossover(const GAGenome& a, const GAGenome& b,
		   GAGenome* c, GAGenome* d){ 
  MixedGenome& mom = (MixedGenome&)a;
  MixedGenome& dad = (MixedGenome&)b;
  int n=0;

  GAGenome::SexualCrossover realcross = mom._greal->sexual();
  GAGenome::SexualCrossover intcross = mom._gint->sexual();

  if(c && d){
    MixedGenome& sis = (MixedGenome&)*c;
    MixedGenome& bro = (MixedGenome&)*d;
    (*realcross)(mom.real_genome(), dad.real_genome(), &sis.real_genome(), &bro.real_genome());
    (*intcross)(mom.int_genome(),dad.int_genome(), &sis.int_genome(), &bro.int_genome());
    n = 2;
    sis._evaluated= gaFalse;
    bro._evaluated= gaFalse;
  }
  else if(c){
    MixedGenome& sis = (MixedGenome&)*c;
    (*realcross)(mom.real_genome(), dad.real_genome(), &sis.real_genome(), 0);
    (*intcross)(mom.int_genome(), dad.int_genome(), &sis.int_genome(), 0);
    n = 1;
    sis._evaluated= gaFalse;
  }
  else if(d){
    MixedGenome& bro = (MixedGenome&)*d;
    (*realcross)(mom.real_genome(), dad.real_genome(), &bro.real_genome(), 0);
    (*intcross)(mom.int_genome(), dad.int_genome(), &bro.int_genome(), 0);
    n = 1;
    bro._evaluated= gaFalse;
  }
  return n;
}










